import asyncio
import http
import re
from datetime import timedelta
from json.decoder import JSONDecodeError
from typing import TYPE_CHECKING
from urllib.parse import parse_qs, quote, unquote, urlencode

from guardpost import Identity

from blacksheep.multipart import parse_multipart
from blacksheep.settings.encodings import encodings_settings
from blacksheep.settings.json import json_settings
from blacksheep.utils.time import utcnow

from .contents import (
    ASGIContent,
    Content,
    multiparts_to_dictionary,
    parse_www_form_urlencoded,
)
from .cookies import Cookie, parse_cookie, split_value, write_cookie_for_response
from .exceptions import BadRequest, BadRequestFormat, FailedRequestError, MessageAborted
from .headers import Headers
from .url import URL, build_absolute_url

if TYPE_CHECKING:
    from blacksheep.sessions import Session

_charset_rx = re.compile(rb"charset=([\w\-]+)", re.I)


def parse_charset(value: bytes):
    m = _charset_rx.search(value)
    if m:
        return m.group(1).decode("ascii")
    return None


async def _read_stream(request):
    async for _ in request.content.stream():
        pass


async def _call_soon(coro):
    task = asyncio.create_task(coro)
    asyncio.get_event_loop().call_soon(task.cancel)
    try:
        return await task
    except asyncio.CancelledError:
        return None


class Message:
    def __init__(self, headers):
        self._raw_headers = headers or []

    @property
    def headers(self) -> Headers:
        key = "_headers"
        if key in self.__dict__:
            return self.__dict__[key]
        self.__dict__[key] = Headers(self._raw_headers)
        return self.__dict__[key]

    def with_content(self, content: Content):
        self.content = content
        return self

    def get_first_header(self, key: bytes):
        key = key.lower()
        for header in self._raw_headers:
            if header[0].lower() == key:
                return header[1]

    def get_headers(self, key: bytes):
        results = []
        key = key.lower()
        for header in self._raw_headers:
            if header[0].lower() == key:
                results.append(header[1])
        return results

    def init_prop(self, name: str, value):
        try:
            getattr(self, name)
        except AttributeError:
            setattr(self, name, value)

    def get_headers_tuples(self, key: bytes):
        results = []
        key = key.lower()
        for header in self._raw_headers:
            if header[0].lower() == key:
                results.append(header)
        return results

    def get_single_header(self, key: bytes):
        results = self.get_headers(key)
        if len(results) > 1:
            raise ValueError("Headers contains more than one header with the given key")
        if len(results) < 1:
            raise ValueError("Headers does not contain one header with the given key")
        return results[0]

    def remove_header(self, key: bytes):
        to_remove = []
        key = key.lower()
        for header in self._raw_headers:
            if header[0].lower() == key:
                to_remove.append(header)
        for header in to_remove:
            self._raw_headers.remove(header)

    def remove_headers(self, headers):
        for header in headers:
            self._raw_headers.remove(header)

    def _has_header(self, key: bytes):
        key = key.lower()
        for existing_key, existing_value in self._raw_headers:
            if existing_key.lower() == key:
                return True
        return False

    def has_header(self, key: bytes):
        return self._has_header(key)

    def _add_header(self, key: bytes, value: bytes):
        self._raw_headers.append((key, value))

    def _add_header_if_missing(self, key: bytes, value: bytes):
        if not self._has_header(key):
            self._raw_headers.append((key, value))

    def add_header(self, key: bytes, value: bytes):
        self._raw_headers.append((key, value))

    def set_header(self, key: bytes, value: bytes):
        self.remove_header(key)
        self._raw_headers.append((key, value))

    def content_type(self):
        if hasattr(self, "content") and self.content and self.content.type:
            return self.content.type
        return self.get_first_header(b"content-type")

    async def read(self):
        if hasattr(self, "content") and self.content:
            return await self.content.read()
        return None

    async def stream(self):
        if hasattr(self, "content") and self.content:
            async for chunk in self.content.stream():
                yield chunk
        else:
            yield None

    async def text(self):
        body = await self.read()
        if body is None:
            return ""
        try:
            return body.decode(self.charset)
        except UnicodeDecodeError as decode_error:
            return encodings_settings.decode(body, decode_error)

    async def form(self):
        content_type_value = self.content_type()
        if not content_type_value:
            return None
        if b"application/x-www-form-urlencoded" in content_type_value:
            text = await self.text()
            return parse_www_form_urlencoded(text)
        if b"multipart/form-data;" in content_type_value:
            body = await self.read()
            return multiparts_to_dictionary(list(parse_multipart(body)))
        return None

    async def multipart(self):
        content_type_value = self.content_type()
        if not content_type_value:
            return None
        if b"multipart/form-data;" in content_type_value:
            body = await self.read()
            return list(parse_multipart(body))
        return None

    def declares_content_type(self, type: bytes):
        content_type = self.content_type()
        if not content_type:
            return False
        if type.lower() in content_type.lower():
            return True
        return False

    def declares_json(self):
        return self.declares_content_type(b"json")

    def declares_xml(self):
        return self.declares_content_type(b"xml")

    async def files(self, name=None):
        if isinstance(name, str):
            name = name.encode("ascii")
        content_type = self.content_type()
        if not content_type or b"multipart/form-data;" not in content_type:
            return []
        data = await self.multipart()
        if name:
            return [part for part in data if part.file_name and part.name == name]
        return [part for part in data if part.file_name]

    async def json(self, loads=json_settings.loads):
        if not self.declares_json():
            return None
        text = await self.text()
        if text is None or text == "":
            return None
        try:
            return loads(text)
        except JSONDecodeError as decode_error:
            content_type = self.content_type()
            if content_type and b"json" in content_type:
                raise BadRequestFormat(
                    f"Declared Content-Type is {content_type.decode()} but "
                    f"the content cannot be parsed as JSON.",
                    decode_error,
                )
            raise BadRequestFormat("Cannot parse content as JSON", decode_error)

    def has_body(self):
        content = getattr(self, "content", None)
        if not content or content.length == 0:
            return False
        return True

    @property
    def charset(self):
        content_type = self.content_type()
        if content_type:
            return parse_charset(content_type) or "utf8"
        return "utf8"


def method_without_body(method: str):
    return method in ("GET", "HEAD", "TRACE")


class Request(Message):
    def __init__(self, method: str, url: bytes, headers):
        _url = URL(url) if url else None
        self._raw_headers = headers or []
        self.method = method
        self._url = _url
        self._session = None
        if _url:
            self._path = _url.path
            self._raw_query = _url.query
        else:
            self._path = None
            self._raw_query = None
        self.scope = None
        self.content: Content | None = None

    # TODO: deprecate the 'identity' property in the future. This requires a breaking
    # change in guardpost, too.
    @property
    def identity(self):
        return self.user

    @identity.setter
    def identity(self, value):
        self.__dict__["_user"] = value

    @property
    def user(self):
        try:
            return self.__dict__["_user"]
        except KeyError:
            self.__dict__["_user"] = Identity()
            return self.__dict__["_user"]

    @user.setter
    def user(self, value):
        self.__dict__["_user"] = value

    @property
    def scheme(self) -> str:
        return self.__dict__.get("scheme") or (
            self.scope.get("scheme", "") if self.scope else ""
        )

    @scheme.setter
    def scheme(self, value: str):
        self.__dict__["scheme"] = value

    @property
    def host(self) -> str:
        if not self.__dict__.get("host"):
            if self._url is not None and self._url.is_absolute:
                self.__dict__["host"] = (
                    self._url.host.decode()
                    if isinstance(self._url.host, bytes)
                    else self._url.host
                )
            else:
                host_header = self.get_first_header(b"host")
                if host_header is None:
                    raise BadRequest("Missing Host header")
                self.__dict__["host"] = host_header.decode()
        return self.__dict__["host"]

    @host.setter
    def host(self, value: str) -> None:
        self.__dict__["host"] = value

    @property
    def path(self) -> str:
        return self._path.decode("utf8") if self._path else ""

    @property
    def base_path(self) -> str:
        try:
            return self.__dict__["base_path"]
        except KeyError:
            try:
                return self.scope.get("root_path", "")
            except AttributeError:
                return ""

    @base_path.setter
    def base_path(self, value: str):
        self.__dict__["base_path"] = value

    @property
    def client_ip(self) -> str:
        if getattr(self, "scope", None) is None:
            return ""
        client = self.scope.get("client", ("", 0))
        return client[0]

    @property
    def original_client_ip(self) -> str:
        if "original_client_ip" in self.__dict__:
            return self.__dict__["original_client_ip"]
        return self.client_ip

    @original_client_ip.setter
    def original_client_ip(self, value: str):
        self.__dict__["original_client_ip"] = value

    @property
    def session(self):
        if self._session is None:
            raise TypeError(
                "A session is not configured for this request, activate "
                "sessions using `app.use_sessions` method."
            )
        return self._session

    @session.setter
    def session(self, value: "Session"):
        self._session = value

    @classmethod
    def incoming(cls, method: str, path: bytes, query: bytes, headers):
        request = cls(method, None, headers)
        request._path = path
        request._raw_query = query
        return request

    @property
    def query(self):
        if self._raw_query:
            return parse_qs(self._raw_query.decode("utf8"))
        return {}

    @query.setter
    def query(self, value):
        raw_query = urlencode(value, True).encode("utf8")
        self._raw_query = raw_query
        self.url = self.url.with_query(raw_query)

    @property
    def url(self):
        if self._url:
            return self._url
        if self._raw_query:
            self._url = URL(self._path + b"?" + self._raw_query)
        else:
            self._url = URL(self._path)
        return self._url

    @url.setter
    def url(self, value):
        if value:
            if isinstance(value, bytes):
                _url = URL(value)
            elif isinstance(value, str):
                _url = URL(value.encode("utf8"))
            elif isinstance(value, URL):
                _url = value
            else:
                raise TypeError("Invalid value type, expected bytes, str, or URL")
        else:
            _url = None
        if _url:
            self._path = _url.path
            self._raw_query = _url.query
        else:
            self._path = None
            self._raw_query = None
        self._url = _url
        self.__dict__["host"] = None
        self.remove_header(b"host")

    def __repr__(self):
        return f"<Request {self.method} {self.url.value.decode()}>"

    @property
    def cookies(self):
        cookies = {}
        cookies_headers = self.get_headers(b"cookie")
        if cookies_headers:
            for header in cookies_headers:
                pairs = header.split(b"; ")
                for fragment in pairs:
                    try:
                        name, value = split_value(fragment, b"=")
                    except ValueError:
                        pass
                    else:
                        cookies[unquote(name.decode())] = unquote(
                            value.rstrip(b"; ").decode()
                        )
        return cookies

    def get_cookie(self, name: str):
        return self.cookies.get(name)

    def set_cookie(self, name: str, value: str):
        new_value = (quote(name) + "=" + quote(value)).encode()
        existing_cookie = self.get_first_header(b"cookie")
        if existing_cookie:
            self.set_header(b"cookie", existing_cookie + b";" + new_value)
        else:
            self._raw_headers.append((b"cookie", new_value))

    @property
    def etag(self):
        return self.get_first_header(b"etag")

    @property
    def if_none_match(self):
        return self.get_first_header(b"if-none-match")

    def expect_100_continue(self):
        value = self.get_first_header(b"expect")
        if value and value.lower() == b"100-continue":
            return True
        return False

    async def is_disconnected(self):
        if not isinstance(self.content, ASGIContent):
            raise TypeError(
                "This method is only supported when a request is bound to "
                "an instance of ASGIContent and to an ASGI "
                "request/response cycle."
            )
        self.init_prop("_is_disconnected", False)
        if self._is_disconnected is True:
            return True
        try:
            await _call_soon(_read_stream(self))
        except MessageAborted:
            self._is_disconnected = True
        return self._is_disconnected


class Response(Message):
    def __init__(self, status: int, headers=None, content: Content = None):
        self._raw_headers = headers or []
        self.status = status
        self.content = content

    def __repr__(self):
        return f"<Response {self.status}>"

    @property
    def cookies(self):
        return self.get_cookies()

    @property
    def reason(self) -> str:
        return http.HTTPStatus(self.status).phrase

    def get_cookies(self):
        cookies = {}
        set_cookies_headers = self.get_headers(b"set-cookie")
        if set_cookies_headers:
            for value in set_cookies_headers:
                cookie = parse_cookie(value)
                cookies[cookie.name] = cookie
        return cookies

    def get_cookie(self, name: str):
        set_cookies_headers = self.get_headers(b"set-cookie")
        if set_cookies_headers:
            for value in set_cookies_headers:
                cookie = parse_cookie(value)
                if cookie.name == name:
                    return cookie
        return None

    def set_cookie(self, cookie: Cookie):
        self._raw_headers.append((b"set-cookie", write_cookie_for_response(cookie)))

    def set_cookies(self, cookies):
        for cookie in cookies:
            self.set_cookie(cookie)

    def unset_cookie(self, name: str):
        self.set_cookie(Cookie(name, "", utcnow() - timedelta(days=365)))

    def remove_cookie(self, name: str):
        to_remove = []
        set_cookies_headers = self.get_headers_tuples(b"set-cookie")
        if set_cookies_headers:
            for value in set_cookies_headers:
                cookie = parse_cookie(value[1])
                if cookie.name == name:
                    to_remove.append(value)
        self.remove_headers(to_remove)

    def is_redirect(self):
        return self.status in {301, 302, 303, 307, 308}

    async def raise_for_status(self):
        if not (200 <= self.status < 300):
            raise FailedRequestError(self.status, await self.text())


def is_cors_request(request: "Request"):
    return bool(request.get_first_header(b"Origin"))


def is_cors_preflight_request(request: "Request"):
    if request.method != "OPTIONS" or not is_cors_request(request):
        return False
    next_request_method = request.get_first_header(b"Access-Control-Request-Method")
    return bool(next_request_method)


def ensure_bytes(value):
    if isinstance(value, str):
        return value.encode()
    if isinstance(value, bytes):
        return value
    raise ValueError("Input value must be bytes or str")


def get_request_absolute_url(request: "Request"):
    if request.url.is_absolute:
        return request.url
    return build_absolute_url(
        ensure_bytes(request.scheme),
        ensure_bytes(request.host),
        ensure_bytes(request.base_path),
        request._path,
    )


def get_absolute_url_to_path(request: "Request", path: str):
    return build_absolute_url(
        ensure_bytes(request.scheme),
        ensure_bytes(request.host),
        ensure_bytes(request.base_path),
        ensure_bytes(path),
    )



import sys
from io import BytesIO
from typing import (
    Any,
    AsyncIterable,
    Callable,
    ClassVar,
    Sequence,
    Type,
)

from blacksheep.common.types import HeadersType, ParamsType
from blacksheep.messages import Request, Response
from blacksheep.server.bindings import ControllerParameter
from blacksheep.server.responses import (
    ContentDispositionType,
    MessageType,
    accepted,
    bad_request,
    created,
    file,
    forbidden,
    html,
    json,
    moved_permanently,
    no_content,
    not_found,
    not_modified,
    permanent_redirect,
    pretty_json,
    redirect,
    see_other,
    status_code,
    temporary_redirect,
    text,
    unauthorized,
    view,
    view_async,
)
from blacksheep.server.routing import (
    RegisteredRoute,
    RouteFilter,
    Router,
)
from blacksheep.server.routing import RoutesRegistry as RoutesRegistry  # noqa
from blacksheep.server.routing import (
    controllers_routes,
    normalize_filters,
)
from blacksheep.utils import AnyStr, ensure_bytes, join_fragments
from blacksheep.utils.meta import all_subclasses, clonefunc

# singleton router used to store initial configuration,
# before the application starts
# this is used as *default* router for controllers, but it can be overridden
# - see for example tests in test_controllers
router = controllers_routes


head = router.head
get = router.get
post = router.post
put = router.put
patch = router.patch
delete = router.delete
trace = router.trace
options = router.options
connect = router.connect
route = router.route
ws = router.ws


class _BaseControllerRegistry:
    """
    Internal class to support skipping the registration of routes defined in base
    controllers. This is used to support defining routes that should be applied only
    in subclasses, combining route prefixes of subclasses.
    """

    _registry = set()

    @classmethod
    def register(cls, controller):
        cls._registry.add(controller)

    @classmethod
    def get_registry(cls):
        """Returns the registry of all controllers marked as base controllers."""
        return cls._registry

    @classmethod
    def is_base_controller(cls, controller):
        """Returns True if the controller is marked as base controller."""
        return controller in cls._registry


def abstract():
    """
    Decorator to mark a controller class as base and indicate that its routes should not
    be registered directly. If applied to a controller class, its routes will be applied
    only in subclasses.

    In the following example, only `/one/hello-world` and `/two/hello-world` routes will
    be applied in the final router (excluding `/hello-world` itself):
    ```python
    @abstract()
    class AppController(Controller):
        @get("/hello-world")
        def index(self):
            return self.text("Hello, World!")

    class ControllerOne(AppController):
        route = "/one"
        # /one/hello-world

    class ControllerTwo(AppController):
        route = "/two"
        # /two/hello-world
    ```
    """

    def class_deco(cls):
        _BaseControllerRegistry.register(cls)
        return cls

    return class_deco


def filters(
    *filters: RouteFilter,
    host: str | None = None,
    headers: HeadersType | None = None,
    params: ParamsType | None = None,
):
    """
    Configures a set of filters for a decorated controller type.
    Filters are applied to all routes defined in the controller.

    ---
    Example: match a "/" route only if the request includes an header "X-Area: Special"

    ```py
    @filters(headers={"X-Area": "Special"})
    class Special(Controller):
        @get("/")
        def special(self):
            ...
    ```
    """

    def class_deco(cls: Type["Controller"]):
        cls._filters_ = normalize_filters(host, headers, params, list(*filters))
        return cls

    return class_deco


class CannotDetermineDefaultViewNameError(RuntimeError):
    def __init__(self):
        super().__init__(
            "Cannot determine the default view name to be used "
            "for the calling function. "
            "Modify your Controller `view()` function call to specify the name"
            " of the view to be used."
        )


class ControllerMeta(type):
    def __init__(cls, name, bases, attr_dict):
        super().__init__(name, bases, attr_dict)

        for value in attr_dict.values():
            if hasattr(value, "route_handler"):
                setattr(value, "controller_type", cls)


class Controller(metaclass=ControllerMeta):
    """Base class for all controllers."""

    _filters_: ClassVar[Sequence[RouteFilter]] = []

    @classmethod
    def route(cls) -> str | None:
        """
        The base route to be used by all request handlers defined on a
        controller type.
        Override this class method in subclasses, to implement base routes.
        """
        return None

    async def on_request(self, request: Request):
        """
        Extensibility point: controllers support executing a function at each
        web request.
        This method is executed before model binding happens: it is possible
        to read values from the request:
        headers are immediately available, and the user can decide to wait for
        the body (e.g. await request.json()).

        Since controllers support dependency injection, it is possible to
        apply logic using dependent services,
        specified in the __init__ constructor.
        """

    async def on_response(self, response: Response):
        """
        Callback called on every response returned using controller's methods.
        """

    def status_code(self, status: int = 200, message: MessageType = None) -> Response:
        """
        Returns a plain response with given status, with optional message;
        sent as plain text or JSON.
        """
        return status_code(status, message)

    def ok(self, message: MessageType = None) -> Response:
        """
        Returns an HTTP 200 OK response, with optional message;
        sent as plain text or JSON.
        """
        return status_code(200, message)

    def created(self, value: Any = None, location: AnyStr = "") -> Response:
        """
        Returns an HTTP 201 Created response, to the given location and with
        optional JSON content.
        """
        return created(value, location)

    def accepted(self, message: MessageType = None) -> Response:
        """
        Returns an HTTP 202 Accepted response, with optional message;
        sent as plain text or JSON.
        """
        return accepted(message)

    def no_content(self) -> Response:
        """
        Returns an HTTP 204 No Content response.
        """
        return no_content()

    def html(self, data, status: int = 200) -> Response:
        """
        Returns a response with text/html content, and given status
        (default HTTP 200 OK).
        """
        return html(data, status)

    def json(self, data, status: int = 200) -> Response:
        """
        Returns a response with application/json content, and given status
        (default HTTP 200 OK).
        """
        return json(data, status)

    def pretty_json(self, data: Any, status: int = 200, indent: int = 4) -> Response:
        """
        Returns a response with indented application/json content,
        and given status (default HTTP 200 OK).
        """
        return pretty_json(data, status=status, indent=indent)

    def text(self, value: str, status: int = 200) -> Response:
        """
        Returns a response with text/plain content,
        and given status (default HTTP 200 OK).
        """
        return text(value, status)

    def moved_permanently(self, location: AnyStr) -> Response:
        """
        Returns an HTTP 301 Moved Permanently response, to the given location.
        """
        return moved_permanently(location)

    def redirect(self, location: AnyStr) -> Response:
        """
        Returns an HTTP 302 Found response (commonly called redirect), to the
        given location.
        """
        return redirect(location)

    def see_other(self, location: AnyStr) -> Response:
        """
        Returns an HTTP 303 See Other response, to the given location.
        """
        return see_other(location)

    def not_modified(self) -> Response:
        """
        Returns an HTTP 304 Not Modified response.
        """
        return not_modified()

    def temporary_redirect(self, location: AnyStr) -> Response:
        """
        Returns an HTTP 307 Temporary Redirect response, to the given location.
        """
        return temporary_redirect(location)

    def permanent_redirect(self, location: AnyStr) -> Response:
        """
        Returns an HTTP 308 Permanent Redirect response, to the given location.
        """
        return permanent_redirect(location)

    def bad_request(self, message: MessageType = None) -> Response:
        """
        Returns an HTTP 400 Bad Request response, with optional message;
        sent as plain text or JSON.
        """
        return bad_request(message)

    def unauthorized(self, message: MessageType = None) -> Response:
        """
        Returns an HTTP 401 Unauthorized response, with optional message;
        sent as plain text or JSON.
        """
        return unauthorized(message)

    def forbidden(self, message: MessageType = None) -> Response:
        """
        Returns an HTTP 403 Forbidden response, with optional message;
        sent as plain text or JSON.
        """
        return forbidden(message)

    def not_found(self, message: MessageType = None) -> Response:
        """
        Returns an HTTP 404 Not Found response, with optional message;
        sent as plain text or JSON.
        """
        return not_found(message)

    def get_default_view_name(self):
        """
        Returns the default view name, to be used by the calling function.
        """
        route_handler_name = self._get_route_handler_name()

        if route_handler_name is None:
            raise CannotDetermineDefaultViewNameError()
        return route_handler_name

    def file(
        self,
        value: Callable[[], AsyncIterable[bytes]] | str | bytes | bytearray | BytesIO,
        content_type: str,
        *,
        file_name: str | None = None,
        content_disposition: ContentDispositionType = ContentDispositionType.ATTACHMENT,
    ) -> Response:
        """
        Returns a binary file response with given content type and optional
        file name, for download (attachment)
        (default HTTP 200 OK). This method supports both call with bytes,
        or a generator yielding chunks.

        Remarks: this method does not handle cache, ETag and HTTP 304 Not Modified
        responses; when handling files it is recommended to handle cache, ETag and
        Not Modified, according to use case.
        """
        return file(
            value,
            content_type,
            content_disposition=content_disposition,
            file_name=file_name,
        )

    def _get_route_handler_name(self):
        """
        Returns the name of the closest route handler in the call stack.

        Note: this function is designed to improve user's experience when
        using the framework.
        It removes the need to explicit the name of the template when using
        the `view()` function.

        It uses sys._getframe, which is a CPython implementation detail and
        is not guaranteed to exist in all implementations of Python.
        https://docs.python.org/3/library/sys.html
        """
        i = 2
        # Note: no need to get the frame for this function and
        # the direct caller;
        while True:
            i += 1
            try:
                fn_meta = sys._getframe(i).f_code
            except AttributeError:
                # NB: if sys._getframe raises attribute error, it means
                # is not supported
                # by the used Python runtime; return None; which in turn
                # will cause an exception.
                return None
            except ValueError:
                break
            fn = getattr(self, fn_meta.co_name, None)

            if not fn:
                continue

            try:
                if fn.route_handler:
                    return fn_meta.co_name
            except AttributeError:
                pass
        return None

    @classmethod
    def class_name(cls) -> str:
        """
        Returns the class name to be used for conventional behavior.
        By default, it returns the lowercase class name.
        """
        return cls.__name__.lower()

    def full_view_name(self, name: str) -> str:
        """
        Returns the full view name for this controller.
        By default, this function concatenates the lowercase class name
        to the view name.

        Therefore, a Home(Controller) will look for templates inside
        /views/home/ folder.
        """
        return f"{self.class_name()}/{name}"

    def view(
        self,
        name: str | None = None,
        model: Any | None = None,
        status: int = 200,
        **kwargs,
    ) -> Response:
        """
        Returns a view rendered synchronously.

        :param name: name of the template (path to the template file,
            optionally without '.html' extension
        :param model: optional model, required to render the template.
        :param status: optional status code for the response, default 200.
        :return: a Response object
        """
        if name is None:
            name = self.get_default_view_name()

        return view(self.full_view_name(name), model, status, **kwargs)

    async def view_async(
        self,
        name: str | None = None,
        model: Any = None,
        status: int = 200,
        **kwargs,
    ) -> Response:
        """
        Returns a view rendered asynchronously.

        :param name: name of the template (path to the template file,
            optionally without '.html' extension
        :param model: optional model, required to render the template.
        :param status: optional status code for the response, default 200.
        :return: a Response object
        """
        if name is None:
            name = self.get_default_view_name()

        return await view_async(self.full_view_name(name), model, status, **kwargs)


class APIController(Controller):
    @classmethod
    def version(cls) -> str | None:
        """
        The version of this api controller. If specified, it is included in
        the base route for this controller.

        Example:
            if version is 'v1', and base route 'cat'; all route handlers
            defined on the controller have prefix:
            /api/v1/cat

            if the class name ends with the version string, the suffix is
            automatically removed from routes, so:
            class CatV2; with version() -> "v2"; produces such routes:
            /api/v2/cat
            And not, ~/api/v2/catv2~!
        """
        return None

    @classmethod
    def route(cls) -> str | None:
        cls_name = cls.class_name()
        cls_version = cls.version() or ""
        if cls_version and cls_name.endswith(cls_version.lower()):
            cls_name = cls_name[: -len(cls_version)]
        return join_fragments("api", cls_version, cls_name)


class ControllersManager:
    """
    This class is used to apply the routes defined in the controllers, and support
    inheritance of routes.
    """

    def prepare_controllers(self, router: Router) -> list[Type]:
        self._unify_controllers(router.controllers_routes)
        controller_types = []
        for route in router.controllers_routes:
            handler = route.handler
            controller_type = getattr(handler, "controller_type")

            sub_classes = [
                sub
                for sub in all_subclasses(controller_type)
                if issubclass(sub, Controller)
            ]
            for sub in sub_classes:
                controller_types.append(sub)

            controller_types.append(controller_type)

            handler.__annotations__["self"] = ControllerParameter[controller_type]
            new_route = router.create_route(
                self.get_controller_handler_pattern(controller_type, route),
                handler,
                controller_type._filters_,
            )
            router.add_route(route.method, new_route)
        return controller_types

    def get_controller_handler_pattern(
        self, controller_type: Type, route: RegisteredRoute
    ) -> bytes:
        """
        Returns the full pattern to be used for a route handler,
        defined as controller method.
        """
        base_route = getattr(controller_type, "route", None)

        if base_route is not None:
            if callable(base_route):
                value = base_route()
            elif isinstance(base_route, (str, bytes)):
                value = base_route
            else:
                raise RuntimeError(
                    f"Invalid controller `route` attribute. "
                    f"Controller `{controller_type.__name__}` "
                    f"has an invalid route attribute: it should "
                    f"be callable, or str, or bytes."
                )

            if value:
                return ensure_bytes(join_fragments(value, route.pattern))
        return ensure_bytes(route.pattern)

    def _handle_controller_subclasses(self, route, handler, controller_type: Type):
        # we need to discover subclasses because the user configures requests handlers
        # on base types and they can be inherited
        sub_classes = [
            sub
            for sub in all_subclasses(controller_type)
            if issubclass(sub, Controller)
        ]
        for sub_class in sub_classes:
            handler_clone = clonefunc(handler)
            handler_clone.controller_type = sub_class
            yield RegisteredRoute(
                method=route.method,
                pattern=route.pattern,
                handler=handler_clone,
            )

    def _unify_controllers(self, controllers_router):
        """
        Unifies the routes of the controllers, to support controllers inheritance.
        This must happen at application start because at this stage all controller types
        are loaded in memory and the routes are registered in the router.
        """
        routes_to_add = []
        routes_to_remove = []
        for route in controllers_router:
            handler = route.handler
            controller_type = getattr(handler, "controller_type")
            # We support skipping the registration of routes defined in base controllers
            if _BaseControllerRegistry.is_base_controller(controller_type):
                routes_to_remove.append(route)

            routes_to_add.extend(
                self._handle_controller_subclasses(route, handler, controller_type)
            )

        for route in routes_to_add:
            controllers_router.routes.append(route)

        for route in routes_to_remove:
            controllers_router.routes.remove(route)



"""
This module provides classes to handle API Keys authentication.
"""

from abc import ABC, abstractmethod
from copy import deepcopy
from enum import Enum
from typing import Literal

from essentials.secrets import Secret
from guardpost import AuthenticationHandler, Identity
from guardpost.errors import InvalidCredentialsError

from blacksheep.messages import Request


class APIKeyLocation(Enum):
    HEADER = "header"
    QUERY = "query"
    COOKIE = "cookie"


APIKeyLocationValue = Literal["header", "query", "cookie"]


class APIKey:
    def __init__(
        self,
        secret: Secret,
        claims: dict | None = None,
        roles: list[str] | None = None,
    ) -> None:
        """
        Creates an instance of API Key for authentication.

        Parameters
        ----------
        secret : Secret
            The secret value of the API key.
        claims : dict | None, optional
            Additional claims to include in the authenticated identity, by default None.
        roles : list[str] | None, optional
            List of roles to assign to the authenticated identity, by default None.

        Raises
        ------
        ValueError
            If both 'secret' and 'secrets' are provided, or if neither is provided.
        """
        self._claims = claims or {}
        self._roles = roles or []
        self._secret = secret

    @property
    def claims(self) -> dict:
        """Returns the claims associated with this API Key."""
        return self._claims

    @property
    def roles(self) -> list[str]:
        """Returns the roles associated with this API Key."""
        return self._roles

    def match(self, secret: Secret | str) -> bool:
        """
        Returns a value indicating if the provided secret matches this API Key secret.
        """
        return self._secret == secret


class APIKeysProvider(ABC):
    """
    Abstract base class for providing API keys dynamically.

    This class defines the interface for API key providers that can be used
    with APIKeyAuthentication to retrieve API keys at runtime. Implementations
    can fetch keys from various sources such as databases, configuration files,
    external services, or in-memory stores.

    Examples
    --------
    >>> class DatabaseAPIKeysProvider(APIKeysProvider):
    ...     async def get_keys(self) -> list[APIKey]:
    ...         # Fetch keys from database…
    ...         api_keys = await self.fetch_api_keys_from_db()
    ...         return [APIKey(key.name, Secret(key.secret, direct_value=True))
    ...                 for key in api_keys]
    """

    @abstractmethod
    async def get_keys(self) -> list[APIKey]:
        """
        Retrieve a list of valid API keys.

        This method should return all currently valid API keys that can be used
        for authentication. The method is called by APIKeyAuthentication during
        the authentication process when no static keys are provided.

        Returns
        -------
        list[APIKey]
            A list of APIKey instances representing all valid keys for authentication.
            An empty list indicates no valid keys are available.

        Notes
        -----
        - This method is called for each authentication attempt, so implementations
          should consider caching strategies for performance if fetching keys is
          expensive.
        - The returned list should only contain currently valid and active API keys.
        """


class APIKeyAuthentication(AuthenticationHandler):
    """
    AuthenticationHandler that uses API Keys for authentication, handling one or more
    API Key secrets. Each secret can be associated to different claims and roles.
    """

    def __init__(
        self,
        *keys: APIKey,
        param_name: str,
        scheme: str = "APIKey",
        location: APIKeyLocationValue | APIKeyLocation = "header",
        keys_provider: APIKeysProvider | None = None,
        description: str | None = None,
    ) -> None:
        """
        Creates a new instance of APIKeyAuthentication.

        Parameters
        ----------
        param_name : str
            Arbitrary name of the API key parameter (e.g., header name, query parameter
            name, or cookie name).
        keys : Optional keys handled by this instance (APIKeys). Use this
            parameter or a keys_provider.
        scheme : str, optional
            The authentication scheme name, by default "APIKey".
        location : APIKeyLocationValue | APIKeyLocation, optional
            Where to look for the API key in the request (header, query, or cookie), by
            default "header".
        keys_provider : APIKeysProvider | None
            An optional provider that can be used to retrieve keys dynamically.
            If not provided, the keys passed as parameters will be used.
        description : str | None
            An optional description for this authentication scheme.
        """
        super().__init__()
        self._scheme = scheme
        self._keys = tuple(keys) if keys else None
        self._keys_provider = keys_provider
        self._param_name = param_name
        self._location = APIKeyLocation(location)
        self.description = description

        if keys and keys_provider:
            raise ValueError("Cannot specify both static keys and a keys_provider")
        elif not keys and keys_provider is None:
            raise ValueError("Either keys or keys_provider must be provided")

    @property
    def scheme(self) -> str:
        """Returns the name of the Authentication Scheme used by this handler."""
        return self._scheme

    @property
    def param_name(self) -> str:
        """Returns the name of the API Key."""
        return self._param_name

    @property
    def location(self) -> APIKeyLocation:
        """Returns the location of the API Key."""
        return self._location

    async def authenticate(self, context: Request) -> Identity | None:
        """
        Tries to authenticate the request using API Keys.
        If authentication succeeds, returns an Identity with the authentication mode set
        to the scheme of the matching API Key. If no matching API Key is found, returns
        None.
        """
        matching_key = await self._match_key(context)

        if matching_key is None:
            # Return None here and do not raise an exception, because the application
            # might be configured with alternative ways to authenticate users.
            return None

        return self._get_identity_for_key(matching_key)

    def _get_identity_for_key(self, key: APIKey) -> Identity:
        """
        Returns an instance of Identity to be used when authentication with a given
        API Key succeeded. Each API Key can be associated with specific roles and
        claims.
        """
        claims = deepcopy(key.claims)
        claims.update({"roles": [role for role in key.roles]})
        return Identity(claims, authentication_mode=self.scheme)

    def _get_input_secret(self, context: Request) -> str | None:
        value = None
        if self.location == APIKeyLocation.HEADER:
            bytes_value = context.get_first_header(self._param_name.encode())
            if bytes_value:
                value = bytes_value.decode()
        elif self.location == APIKeyLocation.QUERY:
            list_value = context.query.get(self._param_name)
            if list_value:
                value = list_value[-1]
        elif self.location == APIKeyLocation.COOKIE:
            value = context.cookies.get(self._param_name)
        else:
            # This should never happen
            raise TypeError("APIKeyLocation not supported.")
        return value

    async def _match_key(self, context: Request) -> APIKey | None:
        """
        Tries to find a matching API Key in the request context.
        Returns the matching API Key if found, otherwise None.

        If the client provides an API Key and it is invalid, an InvalidCredentialsError
        error is raised to keep track of this event and support rate-limiting requests
        from the same client, to prevent brute-forcing.
        """
        keys = self._keys

        if not keys and self._keys_provider is not None:
            keys = await self._keys_provider.get_keys()

        if not keys:
            return None

        input_secret = self._get_input_secret(context)

        if not input_secret:
            return None

        for key in keys:
            if key.match(input_secret):
                return key

        # The client provided an API Key, but it is invalid. This event must be logged,
        # and we must rate-limit this kind of request by client IP.
        raise InvalidCredentialsError(context.original_client_ip)



def extend(obj, cls):
    """
    Applies a mixin to an instance of a class.

    This method is used in those scenarios where opting-in for a feature incurs a
    performance fee, so that said fee is paid only when features are used.
    """
    base_cls = obj.__class__

    # Check if the mixin is already applied
    if cls in base_cls.__mro__:
        return

    # Create a new class that combines the mixin and the original class
    base_cls_name = f"{base_cls.__name__}_{cls.__name__}"
    obj.__class__ = type(base_cls_name, (cls, base_cls), {})



from pydantic import BaseModel

from blacksheep.server.bindings import FromJSON
from blacksheep.testing.helpers import get_example_scope
from blacksheep.testing.messages import MockReceive, MockSend


class CustomClass:

    def __init__(self, obj) -> None:
        self.obj = obj


class Example(BaseModel):
    id: int
    name: str


class ExampleCollection(FromJSON[list[Example]]):

    @staticmethod
    def convert(obj) -> list[Example]:
        return [Example(**item) for item in obj]


class JSONNestedList(FromJSON[list[list[str]]]):

    @staticmethod
    def convert(obj) -> list[list[str]]:
        """
        This function is a custom converter for request bodies. It receives as input the
        already JSON-parsed Python object (not the raw JSON string), and can return any
        custom Python object as needed for your application logic.
        """
        # return the object as-is
        return obj


async def _post_scenario(app, request_body):
    await app(
        get_example_scope(
            "POST",
            "/",
            [
                (b"content-type", b"application/json"),
                (b"content-length", str(len(request_body)).encode()),
            ],
        ),
        MockReceive([request_body]),
        MockSend(),
    )
    assert app.response.status == 204


async def test_custom_converter_in_class_definition_1(app):
    request_body = b'[["one","two"],["three"]]'
    expected_result = [["one", "two"], ["three"]]

    @app.router.post("/")
    async def home(item: JSONNestedList):
        assert item is not None
        value = item.value
        assert value == expected_result

    await _post_scenario(app, request_body)


async def test_custom_converter_in_class_definition_2(app):
    request_body = b'[{"id": 1, "name": "Hello"}, {"id": 2, "name": "World"}]'
    expected_result = [Example(id=1, name="Hello"), Example(id=2, name="World")]

    @app.router.post("/")
    async def home(item: FromJSON[list[Example]]):
        assert item is not None
        value = item.value
        assert value == expected_result

    await _post_scenario(app, request_body)


async def test_custom_class_handling(app):
    request_body = b'[["one","two"],["three"]]'
    expected_result = CustomClass([["one", "two"], ["three"]])

    @app.router.post("/")
    async def home(item: FromJSON[CustomClass]):
        assert item is not None
        assert isinstance(expected_result, CustomClass)
        assert expected_result.obj == expected_result.obj

    await _post_scenario(app, request_body)



import http
import inspect
import logging
from collections import UserDict

from .contents cimport Content, TextContent
from .exceptions cimport (
    BadRequest,
    HTTPException,
    InternalServerError,
    InvalidExceptionHandler,
)
from .messages cimport Request, Response

from .utils import get_class_instance_hierarchy

# Better support for Pydantic
try:
    from pydantic import ValidationError
except ImportError:
    ValidationError = None


class ExceptionHandlersDict(UserDict):

    def __setitem__(self, key, item) -> None:
        if not inspect.iscoroutinefunction(item):
            raise InvalidExceptionHandler()
        signature = inspect.Signature.from_callable(item)
        if len(signature.parameters) != 3 and not any(
            param
            for param in signature.parameters
            if signature.parameters[param].kind == 2
        ):
            raise InvalidExceptionHandler()
        return super().__setitem__(key, item)


async def handle_not_found(app, Request request, HTTPException http_exception):
    """Default Not Found handler, returns a simple 404 response."""
    return Response(404, content=TextContent("Resource not found"))


async def handle_internal_server_error(app, Request request, Exception exception):
    """Default Internal Server Error handler, returns a simple 500 response."""
    # Intentionally without details!
    return Response(500, content=TextContent("Internal Server Error"))


async def handle_bad_request(app, Request request, HTTPException http_exception):
    # supports for pydantic ValidationError with json() method
    if http_exception.__context__ is not None and callable(getattr(http_exception.__context__, "json", None)):
        return Response(http_exception.status, content=Content(b"application/json", http_exception.__context__.json().encode("utf8")))

    return Response(400, content=TextContent(f'Bad Request: {str(http_exception)}'))


async def _default_pydantic_validation_error_handler(app, Request request, Exception error):
    return Response(400, content=Content(b"application/json", error.json(indent=4).encode("utf-8")))


async def common_http_exception_handler(app, Request request, HTTPException http_exception):
    return Response(http_exception.status, content=TextContent(http.HTTPStatus(http_exception.status).phrase))


def get_logger():
    logger = logging.getLogger("blacksheep.server")
    logger.setLevel(logging.INFO)
    return logger


cdef class BaseApplication:

    def __init__(self, bint show_error_details, object router):
        self.router = router
        self.exceptions_handlers = self.init_exceptions_handlers()
        self.show_error_details = show_error_details
        self.logger = get_logger()

    def init_exceptions_handlers(self):
        default_handlers = ExceptionHandlersDict({
            404: handle_not_found,
            400: handle_bad_request
        })
        if ValidationError is not None:
            default_handlers[ValidationError] = _default_pydantic_validation_error_handler
        return default_handlers

    async def log_unhandled_exc(self, request, exc):
        self.logger.error(
            "Unhandled exception - \"%s %s\"",
            request.method,
            request.url.value.decode(),
            exc_info=exc
        )

    async def log_handled_exc(self, request, exc):
        if isinstance(exc, HTTPException):
            self.logger.info(
                "HTTP %s - \"%s %s\". %s",
                exc.status,
                request.method,
                request.url.value.decode(),
                str(exc)
            )
        else:
            self.logger.info(
                "Handled error: \"%s %s\". %s",
                request.method,
                request.url.value.decode(),
                str(exc)
            )

    async def handle(self, Request request):
        cdef object route
        cdef Response response

        route = self.router.get_match(request)

        if not route:
            # This is intentional. We should not use user-defined not found exception
            # handlers here because middlewares are not executed. The main router should
            # always have a fallback route configured.
            return Response(404)

        request.route_values = route.values
        try:
            response = await route.handler(request)
        except Exception as exc:
            response = await self.handle_request_handler_exception(request, exc)
        return response or Response(204)

    async def handle_request_handler_exception(self, request, exc):
        if isinstance(exc, HTTPException):
            await self.log_handled_exc(request, exc)
            return await self.handle_http_exception(request, exc)

        if self.is_handled_exception(exc):
            await self.log_handled_exc(request, exc)
        else:
            await self.log_unhandled_exc(request, exc)

        return await self.handle_exception(request, exc)

    cpdef object get_http_exception_handler(self, HTTPException http_exception):
        # Try getting HTTP exception handler by type first, supporting
        # base classes up to a certain point (HTTPException)
        handler = self.get_exception_handler(http_exception, stop_at=HTTPException)
        if handler:
            return handler
        # Try getting HTTP exception handler by HTTP error status code
        return self.exceptions_handlers.get(
            http_exception.status, common_http_exception_handler
        )

    cdef bint is_handled_exception(self, Exception exception):
        for class_type in get_class_instance_hierarchy(exception):
            if class_type in self.exceptions_handlers:
                return True
        return False

    cdef object get_exception_handler(self, Exception exception, type stop_at):
        for class_type in get_class_instance_hierarchy(exception):
            if stop_at is not None and stop_at is class_type:
                return None
            if class_type in self.exceptions_handlers:
                return self.exceptions_handlers[class_type]

        return None

    async def handle_internal_server_error(self, Request request, Exception exc):
        """
        Handle an unhandled exception. If an exception handler is defined for
        InternalServerError or status 500, it is used.
        """
        if self.show_error_details:
            return self.server_error_details_handler.produce_response(request, exc)

        # We want to hide exception details, and possibly use a user-defined
        # handler for this.
        error = InternalServerError(exc)
        internal_server_error_handler = self.get_http_exception_handler(error)

        try:
            return await internal_server_error_handler(self, request, error)
        except Exception:
            self.logger.exception(
                "An exception occurred while trying to apply the configured "
                "Internal Server Error handler!"
            )
        return Response(500, content=TextContent("Internal Server Error"))

    async def _apply_exception_handler(self, Request request, Exception exc, object exception_handler):
        try:
            return await exception_handler(self, request, exc)
        except Exception as server_ex:
            # If the exception happens in the user-defined exception handler,
            # we need to fallback to the default handlers.
            self.logger.error("Unhandled exception in exception_handler: %s", exception_handler.__name__)
            if self.show_error_details:
                return self.server_error_details_handler.produce_response(request, exc)

            return await handle_internal_server_error(self, request, server_ex)

    async def handle_http_exception(self, Request request, HTTPException http_exception):
        exception_handler = self.get_http_exception_handler(http_exception)
        if exception_handler:
            return await self._apply_exception_handler(request, http_exception, exception_handler)

        return await self.handle_exception(request, http_exception)

    async def handle_exception(self, request, exc):
        exception_handler = self.get_exception_handler(exc, None)
        if exception_handler:
            return await self._apply_exception_handler(request, exc, exception_handler)

        return await self.handle_internal_server_error(request, exc)



from urllib.parse import urlparse


class InvalidURL(Exception):
    def __init__(self, message: str):
        super().__init__(message)


def valid_schema(schema):
    if schema and schema != "https" and schema != "http":
        raise InvalidURL(f"Expected http or https schema; got instead {schema}")


class URL:
    def __init__(self, value: bytes):
        if not value:
            raise InvalidURL("Input empty or null.")
        try:
            if value and value[0] == 46:  # ord('.') == 46
                value = b"/" + value
            s = value.decode()
            parsed = urlparse(s)
        except Exception:
            raise InvalidURL(f"The value cannot be parsed as URL ({value.decode()})")
        schema = parsed.scheme
        valid_schema(schema)
        self.value = value or b""
        self.schema = schema.encode() if schema else None
        self.host = parsed.hostname.encode() if parsed.hostname else None
        self.port = parsed.port or 0
        self.path = parsed.path.encode() or b""
        self.query = parsed.query.encode() if parsed.query else None
        self.fragment = parsed.fragment.encode() if parsed.fragment else None
        self.is_absolute = bool(parsed.scheme)

    def __repr__(self):
        return f"<URL {self.value}>"

    def __str__(self):
        return self.value.decode()

    def join(self, other):
        if isinstance(other, bytes):
            other = URL(other)
        if other.is_absolute:
            raise ValueError(
                f"Cannot concatenate to an absolute URL ({self.value} + {other.value})"
            )
        if self.query or self.fragment:
            raise ValueError(
                "Cannot concatenate a URL with query or fragment to another URL portion"
            )
        first_part = self.value
        other_part = other.value
        if first_part and other_part and first_part[-1] == 47 and other_part[0] == 47:
            return URL(first_part[:-1] + other_part)
        return URL(first_part + other_part)

    def base_url(self):
        if not self.is_absolute:
            raise ValueError(
                "This URL is relative. Cannot extract a base URL (without path)."
            )
        base_url = f"{self.schema.decode()}://{self.host.decode()}"  # type: ignore
        if self.port != 0:
            if (self.schema == b"http" and self.port != 80) or (
                self.schema == b"https" and self.port != 443
            ):
                base_url = f"{base_url}:{self.port}"
        return URL(base_url.encode())

    def with_host(self, host: bytes):
        if not self.is_absolute:
            raise TypeError("Cannot generate a URL from a partial URL")
        query = b"?" + self.query if self.query else b""
        fragment = b"#" + self.fragment if self.fragment else b""
        url = f"{self.schema}://{host.decode()}{self.path}{query.decode()}{fragment.decode()}"
        return URL(url.encode())

    def with_query(self, query: bytes):
        query = b"?" + query if query else b""
        fragment = b"#" + self.fragment if self.fragment else b""
        if self.is_absolute:
            url = f"{self.schema}://{self.host}{self.path}{query.decode()}{fragment.decode()}"
            return URL(url.encode())
        url = f"{self.path}{query.decode()}{fragment.decode()}"
        return URL(url.encode())

    def with_scheme(self, schema: bytes):
        s = self.value.decode()
        if isinstance(schema, bytes):
            schema_str = schema.decode()
        else:
            schema_str = schema
        idx = s.find("://")
        if idx == -1:
            raise InvalidURL("Malformed URL for scheme replacement")
        # Ensure s is str, schema_str is str
        new_url = str(schema_str) + str(s[idx:])
        return URL(new_url.encode())

    def __add__(self, other):
        if isinstance(other, bytes):
            return self.join(URL(other))
        if isinstance(other, URL):
            return self.join(other)
        return NotImplemented

    def __eq__(self, other):
        if isinstance(other, URL):
            return self.value == other.value
        return NotImplemented


def build_absolute_url(scheme: bytes, host: bytes, base_path: bytes, path: bytes):
    scheme_str = scheme.decode() if isinstance(scheme, bytes) else scheme
    valid_schema(scheme_str)
    url = (
        str(scheme_str)
        + "://"
        + host.decode()
        + ("/" if base_path else "")
        + base_path.lstrip(b"/").rstrip(b"/").decode()
        + ("/" if path else "")
        + path.lstrip(b"/").decode()
    )
    return URL(url.encode())


def join_prefix(prefix: str, path: str) -> str:
    if not prefix:
        return path
    if not prefix.startswith("/"):
        prefix = "/" + prefix
    if not path:
        return prefix + "/"
    if prefix[-1] == "/" and path[0] == "/":
        return prefix + path[1:]
    if prefix[-1] != "/" and path[0] != "/":
        return prefix + "/" + path
    return prefix + path



from dataclasses import dataclass

from blacksheep.server.responses import ok
from blacksheep.server.routing import Router
from blacksheep.testing.helpers import get_example_scope
from blacksheep.testing.messages import MockReceive, MockSend
from tests.utils.application import FakeApplication


@dataclass
class Cat:
    id: int
    name: str


async def test_application_encoding_error_1():
    app = FakeApplication(router=Router())

    @app.router.post("/")
    def home(data: Cat):
        return ok(data)

    # Simulate a request where the client declares a wrong encoding
    # the payload is encoded using ISO-8859-1 but the client declares UTF-8
    scope = get_example_scope(
        "POST",
        "/",
        [(b"Content-Type", b"Content-Type: application/json; charset=UTF-8")],
    )

    await app(
        scope,
        MockReceive(['{"id": 1, "name": "Café"}'.encode("ISO-8859-1")]),
        MockSend(),
    )

    response = app.response
    # Response status is Bad Request 400
    assert response is not None
    assert response.status == 400
    # The response body contains useful information
    text = await response.text()
    assert "Cannot decode the request content using: utf-8." in text


async def test_application_encoding_correct_1():
    app = FakeApplication(router=Router())

    @app.router.post("/")
    def home(data: Cat):
        return ok(data)

    # Simulate a request where the client declares properly an encoding different than
    # UTF-8
    scope = get_example_scope(
        "POST",
        "/",
        [(b"Content-Type", b"Content-Type: application/json; charset=ISO-8859-1")],
    )

    await app(
        scope,
        MockReceive(['{"id": 1, "name": "Café"}'.encode("ISO-8859-1")]),
        MockSend(),
    )

    response = app.response

    assert response is not None
    assert response.status == 200
    # The response body contains useful information
    text = await response.text()
    assert '{"id":1,"name":"Café"}' == text



import http
import re

from .contents import Content, StreamedContent
from .cookies import Cookie, write_cookie_for_response
from .messages import Request, Response

MAX_RESPONSE_CHUNK_SIZE = 61440  # 64kb


# Header writing utilities
def write_header(header):
    return header[0] + b": " + header[1] + b"\r\n"


def write_headers(headers):
    value = bytearray()
    for header in headers:
        value.extend(write_header(header))
    return bytes(value)


def extend_data_with_headers(headers, data: bytearray):
    for header in headers:
        data.extend(write_header(header))


def _get_status_line(status_code: int):
    try:
        return (
            b"HTTP/1.1 "
            + str(status_code).encode()
            + b" "
            + http.HTTPStatus(status_code).phrase.encode()
            + b"\r\n"
        )
    except ValueError:
        return b"HTTP/1.1 " + str(status_code).encode() + b"\r\n"


STATUS_LINES = {
    status_code: _get_status_line(status_code) for status_code in range(100, 600)
}


def get_status_line(status: int):
    return STATUS_LINES[status]


def write_request_method(request: Request):
    return request.method.encode()


def write_request_uri(request: Request):
    url = request.url
    p = url.path or b"/"
    if url.query:
        return p + b"?" + url.query
    return p


def ensure_host_header(request: Request):
    if request.url.host:
        request._add_header_if_missing(b"host", request.url.host)


def should_use_chunked_encoding(content: Content):
    return content.length < 0


def set_headers_for_response_content(message: Response):
    content = message.content
    if not content:
        message._add_header(b"content-length", b"0")
        return
    message._add_header(b"content-type", content.type or b"application/octet-stream")
    if should_use_chunked_encoding(content):
        message._add_header(b"transfer-encoding", b"chunked")
    else:
        message._add_header(b"content-length", str(content.length).encode())


def set_headers_for_content(message):
    content = message.content
    if not content:
        message._add_header_if_missing(b"content-length", b"0")
        return
    message._add_header_if_missing(
        b"content-type", content.type or b"application/octet-stream"
    )
    if should_use_chunked_encoding(content):
        message._add_header_if_missing(b"transfer-encoding", b"chunked")
    else:
        message._add_header_if_missing(b"content-length", str(content.length).encode())


def write_response_cookie(cookie: Cookie):
    return write_cookie_for_response(cookie)


async def write_chunks(http_content: Content):
    async for chunk in http_content.get_parts():
        yield (hex(len(chunk)))[2:].encode() + b"\r\n" + chunk + b"\r\n"
    yield b"0\r\n\r\n"


def is_small_response(response: Response):
    content = response.content
    if not content:
        return True
    if (
        content.length > 0
        and content.length < MAX_RESPONSE_CHUNK_SIZE
        and content.body is not None
    ):
        return True
    return False


def is_small_request(request: Request):
    content = request.content
    if not content:
        return True
    if (
        content.length > 0
        and content.length < MAX_RESPONSE_CHUNK_SIZE
        and content.body is not None
    ):
        return True
    return False


def request_has_body(request: Request):
    content = request.content
    if not content or content.length == 0:
        return False
    return True


def write_request_without_body(request: Request):
    data = bytearray()
    data.extend(
        write_request_method(request)
        + b" "
        + write_request_uri(request)
        + b" HTTP/1.1\r\n"
    )
    ensure_host_header(request)
    extend_data_with_headers(request._raw_headers, data)
    data.extend(b"\r\n")
    return bytes(data)


def write_small_request(request: Request):
    data = bytearray()
    data.extend(
        write_request_method(request)
        + b" "
        + write_request_uri(request)
        + b" HTTP/1.1\r\n"
    )
    ensure_host_header(request)
    set_headers_for_content(request)
    extend_data_with_headers(request._raw_headers, data)
    data.extend(b"\r\n")
    if request.content:
        data.extend(request.content.body)
    return bytes(data)


def write_small_response(response: Response):
    data = bytearray()
    data.extend(STATUS_LINES[response.status])
    set_headers_for_content(response)
    extend_data_with_headers(response._raw_headers, data)
    data.extend(b"\r\n")
    if response.content:
        data.extend(response.content.body)
    return bytes(data)


def py_write_small_response(response: Response):
    return write_small_response(response)


def py_write_small_request(request: Request):
    return write_small_request(request)


async def write_request_body_only(request: Request):
    content = request.content
    if content:
        if should_use_chunked_encoding(content):
            async for chunk in write_chunks(content):
                yield chunk
        elif isinstance(content, StreamedContent):
            async for chunk in content.get_parts():
                yield chunk
        else:
            data = content.body
            if content.length > MAX_RESPONSE_CHUNK_SIZE:
                for chunk in get_chunks(data):
                    yield chunk
            else:
                yield data
    else:
        raise ValueError("Missing request content")


async def write_request(request: Request):
    ensure_host_header(request)
    set_headers_for_content(request)
    yield write_request_method(request) + b" " + write_request_uri(
        request
    ) + b" HTTP/1.1\r\n" + write_headers(request._raw_headers) + b"\r\n"
    content = request.content
    if content:
        if should_use_chunked_encoding(content):
            async for chunk in write_chunks(content):
                yield chunk
        elif isinstance(content, StreamedContent):
            async for chunk in content.get_parts():
                yield chunk
        else:
            data = content.body
            if content.length > MAX_RESPONSE_CHUNK_SIZE:
                for chunk in get_chunks(data):
                    yield chunk
            else:
                yield data


def get_chunks(data: bytes):
    for i in range(0, len(data), MAX_RESPONSE_CHUNK_SIZE):
        yield data[i : i + MAX_RESPONSE_CHUNK_SIZE]
    yield b""


async def write_response_content(response: Response):
    content = response.content
    if content:
        if should_use_chunked_encoding(content):
            async for chunk in write_chunks(content):
                yield chunk
        elif isinstance(content, StreamedContent):
            async for chunk in content.get_parts():
                yield chunk
        else:
            data = content.body
            if content.length > MAX_RESPONSE_CHUNK_SIZE:
                for chunk in get_chunks(data):
                    yield chunk
            else:
                yield data


async def write_response(response: Response):
    set_headers_for_content(response)
    yield STATUS_LINES[response.status] + write_headers(response._raw_headers) + b"\r\n"
    async for chunk in write_response_content(response):
        yield chunk


async def send_asgi_response(response: Response, send):
    content = response.content
    set_headers_for_response_content(response)
    await send(
        {
            "type": "http.response.start",
            "status": response.status,
            "headers": response._raw_headers,
        }
    )
    if content:
        if content.length < 0 or isinstance(content, StreamedContent):
            closing_chunk = False
            async for chunk in content.get_parts():
                if not chunk:
                    closing_chunk = True
                await send(
                    {
                        "type": "http.response.body",
                        "body": chunk,
                        "more_body": bool(chunk),
                    }
                )
            if not closing_chunk:
                await send(
                    {"type": "http.response.body", "body": b"", "more_body": False}
                )
        else:
            if content.length > MAX_RESPONSE_CHUNK_SIZE:
                for chunk in get_chunks(content.body):
                    await send(
                        {
                            "type": "http.response.body",
                            "body": chunk,
                            "more_body": bool(chunk),
                        }
                    )
            else:
                await send(
                    {
                        "type": "http.response.body",
                        "body": content.body,
                        "more_body": False,
                    }
                )
    else:
        await send({"type": "http.response.body", "body": b""})


_NEW_LINES_RX = re.compile("\r\n|\n")


def write_sse(event):
    value = bytearray()
    if event.id:
        value.extend(b"id: " + _NEW_LINES_RX.sub("", event.id).encode("utf8") + b"\n")
    if event.comment:
        for part in _NEW_LINES_RX.split(event.comment):
            value.extend(b": " + part.encode("utf8") + b"\n")
    if event.event:
        value.extend(
            b"event: " + _NEW_LINES_RX.sub("", event.event).encode("utf8") + b"\n"
        )
    if event.data:
        value.extend(b"data: " + event.write_data().encode("utf8") + b"\n")
    if event.retry > -1:
        value.extend(b"retry: " + str(event.retry).encode() + b"\n")
    value.extend(b"\n")
    return bytes(value)



from abc import ABC, abstractmethod


class Decoder(ABC):
    """
    Abstract base class for byte sequence decoders.

    Implementations of this class provide a strategy for decoding bytes into
    strings, used when a UnicodeDecodeError occurs during standard
    decoding. Subclasses must implement the `decode` method, which receives
    the bytes to decode and the original UnicodeDecodeError.

    Methods:
        decode(value: bytes, decode_error: UnicodeDecodeError) -> str:
            Attempts to decode the given bytes. Should raise the provided
            decode_error if decoding is not possible.
    """

    @abstractmethod
    def decode(self, value: bytes, decode_error: UnicodeDecodeError) -> str: ...


class NoopDecoder(Decoder):
    """
    A decoder implementation that does not attempt to decode input bytes.

    This class always raises the provided UnicodeDecodeError when its decode
    method is called. It can be used to disable automatic encoding detection
    and force strict decoding behavior, ensuring that decoding errors are
    not silently handled or guessed.

    Methods:
        decode(value: bytes, decode_error: UnicodeDecodeError) -> str:
            Always raises the provided decode_error.
    """

    def decode(self, value: bytes, decode_error: UnicodeDecodeError) -> str:
        raise decode_error


class EncodingsSettings:
    """
    Manages the decoding strategy for byte sequences in the application.

    EncodingsSettings allows configuring which Decoder implementation is used
    to decode bytes when a UnicodeDecodeError occurs. By default, it uses
    NoopDecoder, which does not attempt to detect the encoding and re-raises the
    UnicodeDecodeError for further processing. The decoder can be replaced at runtime
    using the `use` method.

    Methods:
        use(decoder: Decoder) -> None:
            Sets the decoder to be used for decoding operations.

        decode(value: bytes, decode_error: UnicodeDecodeError) -> str:
            Decodes the given bytes using the configured decoder. If decoding fails,
            the provided UnicodeDecodeError is raised or handled according to the
            decoder.
    """

    def __init__(self) -> None:
        self._decoder = NoopDecoder()

    def use(self, decoder: Decoder) -> None:
        self._decoder = decoder

    def decode(self, value: bytes, decode_error: UnicodeDecodeError) -> str:
        return self._decoder.decode(value, decode_error)


encodings_settings = EncodingsSettings()



from datetime import datetime
from email.utils import parsedate_to_datetime
from enum import Enum
from urllib.parse import quote, unquote


class CookieSameSiteMode(Enum):
    UNDEFINED = 0
    STRICT = 1
    LAX = 2
    NONE = 3


def datetime_to_cookie_format(value: datetime) -> bytes:
    return value.strftime("%a, %d %b %Y %H:%M:%S GMT").encode()


def datetime_from_cookie_format(value: bytes) -> datetime:
    return parsedate_to_datetime(value.decode()).replace(tzinfo=None)


class CookieError(Exception):
    def __init__(self, message: str):
        super().__init__(message)


class CookieValueExceedsMaximumLength(CookieError):
    def __init__(self):
        super().__init__(
            "The length of the cookie value exceeds the maximum "
            "length of 4096 bytes, and it would be ignored or truncated "
            "by clients. See: https://tools.ietf.org/html/rfc6265#section-6.1"
        )


class Cookie:
    def __init__(
        self,
        name: str,
        value: str,
        expires: datetime | None = None,
        domain: str | None = None,
        path: str | None = None,
        http_only: bool = False,
        secure: bool = False,
        max_age: int = -1,
        same_site: CookieSameSiteMode = CookieSameSiteMode.UNDEFINED,
    ):
        self.name = name
        self.value = value
        self.expires = expires
        self.domain = domain
        self.path = path
        self.http_only = http_only
        self.secure = secure
        self.max_age = max_age
        self.same_site = same_site

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        if not value:
            raise ValueError("A cookie name is required")
        self._name = value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        if value and len(value.encode()) > 4096:
            raise CookieValueExceedsMaximumLength()
        self._value = value

    def clone(self):
        return Cookie(
            self.name,
            self.value,
            self.expires,
            self.domain,
            self.path,
            self.http_only,
            self.secure,
            self.max_age,
            self.same_site,
        )

    def __eq__(self, other):
        if isinstance(other, str):
            return other == self.value
        if isinstance(other, bytes):
            return other.decode() == self.value
        if isinstance(other, Cookie):
            return other.name == self.name and other.value == self.value
        return NotImplemented

    def __repr__(self):
        return f"<Cookie {self.name}: {self.value}>"


def split_value(raw_value: bytes, separator: bytes):
    try:
        index = raw_value.index(separator)
    except ValueError:
        # this is the situation of flags, e.g. httponly and secure
        return b"", raw_value
    return raw_value[:index], raw_value[index + 1 :]


def same_site_mode_from_bytes(raw_value: bytes) -> CookieSameSiteMode:
    if not raw_value:
        return CookieSameSiteMode.UNDEFINED
    raw_value_lower = raw_value.lower()
    if raw_value_lower == b"strict":
        return CookieSameSiteMode.STRICT
    if raw_value_lower == b"lax":
        return CookieSameSiteMode.LAX
    if raw_value_lower == b"none":
        return CookieSameSiteMode.NONE
    return CookieSameSiteMode.UNDEFINED


def parse_cookie(raw_value: bytes) -> Cookie:
    eq = b"="
    parts = raw_value.split(b"; ")
    if len(parts) == 0:
        # only name=value pair
        try:
            name, value = split_value(raw_value, eq)
        except ValueError:
            raise ValueError(f"Invalid name=value fragment: {parts[0]}")
        else:
            return Cookie(name.decode(), value.decode())
    if len(parts) == 1:
        # some set a cookie with a separator without space
        parts = raw_value.split(b";")
    try:
        name, value = split_value(parts[0], eq)
    except ValueError:
        raise ValueError(f"Invalid name=value fragment: {parts[0]}")
    if b" " in value and value.startswith(b'"'):
        value = value.strip(b'"')
    expires = None
    domain = None
    path = None
    http_only = False
    secure = False
    max_age = -1
    same_site = None
    for part in parts:
        if eq in part:
            k, v = split_value(part, eq)
            lower_k = k.lower()
            if lower_k == b"expires":
                expires = v
            elif lower_k == b"domain":
                domain = v
            elif lower_k == b"path":
                path = v
            elif lower_k == b"max-age":
                max_age = int(v)
            elif lower_k == b"samesite":
                same_site = v
        else:
            lower_part = part.lower()
            if lower_part == b"httponly":
                http_only = True
            if lower_part == b"secure":
                secure = True
    return Cookie(
        unquote(name.decode()),
        unquote(value.decode()),
        datetime_from_cookie_format(expires) if expires else None,
        domain.decode() if domain else None,
        path.decode() if path else None,
        http_only,
        secure,
        max_age,
        same_site_mode_from_bytes(same_site),
    )


def write_cookie_for_response(cookie: Cookie) -> bytes:
    parts = []
    parts.append(quote(cookie.name).encode() + b"=" + quote(cookie.value).encode())
    if cookie.expires:
        parts.append(b"Expires=" + datetime_to_cookie_format(cookie.expires))
    if cookie.max_age > -1:
        parts.append(b"Max-Age=" + str(cookie.max_age).encode())
    if cookie.domain:
        parts.append(b"Domain=" + cookie.domain.encode())
    if cookie.path:
        parts.append(b"Path=" + cookie.path.encode())
    if cookie.http_only:
        parts.append(b"HttpOnly")
    if (
        cookie.secure
        or cookie.same_site == CookieSameSiteMode.STRICT
        or cookie.same_site == CookieSameSiteMode.NONE
    ):
        parts.append(b"Secure")
    if cookie.same_site == CookieSameSiteMode.STRICT:
        parts.append(b"SameSite=Strict")
    if cookie.same_site == CookieSameSiteMode.LAX:
        parts.append(b"SameSite=Lax")
    if cookie.same_site == CookieSameSiteMode.NONE:
        parts.append(b"SameSite=None")
    return b"; ".join(parts)



from blacksheep import Application, Router, html
from blacksheep.server.errors import ServerErrorDetailsHandler


def _get_response_without_details(request, info, error_page_template: str):
    content = error_page_template.format_map(
        {
            "info": info,
            "exctype": "",
            "excmessage": "",
            "method": request.method,
            "path": request.url.value.decode(),
            "full_url": "",
        }
    )

    return html(content, status=503)


def get_diagnostic_app(exc: Exception, match: str = "/*") -> Application:
    """
    Returns a fallback application, to help diagnosing start-up errors.

    Example:

        def get_app():
            try:
                # ... your code that configures the application object
                return configure_application()
            except Exception as exc:
                return get_diagnostic_app(exc)


        app = get_app()
    """
    router = Router()
    error_details_handler = ServerErrorDetailsHandler()

    app = Application(router=router)

    @router.get(match)
    async def diagnostic_home(request):
        if app.show_error_details:
            response = error_details_handler.produce_response(request, exc)
            response.status = 503
            return response
        return _get_response_without_details(
            request,
            (
                "The application failed to start. Error details are hidden for security"
                " reasons. To display temporarily error details and investigate the "
                "issue, configure temporarily the environment to display error details."
                "APP_SHOW_ERROR_DETAILS=1"
            ),
            error_details_handler._error_page_template,
        )

    return app



from typing import Awaitable, Callable

from blacksheep import Response
from blacksheep.server.responses import moved_permanently


def default_trailing_slash_exclude(path: str) -> bool:
    return "/api/" in path


def get_trailing_slash_middleware(
    exclude: Callable[[str], bool] | None = None,
) -> Callable[..., Awaitable[Response]]:
    """
    Returns a middleware that redirects requests that do not end with a trailing slash
    to the same URL with a trailing slash, with a HTTP 301 Moved Permanently response.
    This is useful for endpoints that serve HTML documents, to ensure that relative
    URLs in the response body are correctly resolved.
    To filter certain requests from being redirected, pass a callable that returns
    True if the request should be excluded from redirection, by path.
    The default exclude function excludes all requests whose path contains "/api/".
    """
    if exclude is None:
        exclude = default_trailing_slash_exclude

    async def trailing_slash_middleware(request, handler):
        path = request.path

        if exclude and exclude(path):
            return await handler(request)

        if not path.endswith("/") and "." not in path.split("/")[-1]:
            return moved_permanently(f"/{path.strip('/')}/")
        return await handler(request)

    return trailing_slash_middleware



import logging


def get_logger():
    """
    Returns a "blacksheep.sessions" logger.
    """
    logger = logging.getLogger("blacksheep.sessions")
    logger.setLevel(logging.INFO)
    return logger



"""
This module offers methods to return file paths for resources. Its original purpose is
to provide contents of static files stored for conveniency in the blacksheep.server.res
package.
"""

from importlib.resources import files


def get_resource_file_path(anchor, file_name: str) -> str:
    return str(files(anchor) / file_name)


def get_resource_file_content(file_name: str) -> str:
    with open(
        get_resource_file_path("blacksheep.server.res", file_name),
        mode="rt",
    ) as source:
        return source.read()



from abc import ABC, abstractmethod
from typing import Any, Mapping

from blacksheep.messages import Request, Response


class Session:
    """
    Represents a session for storing and managing user-specific data across requests.

    The Session class provides a dictionary-like interface for storing key-value pairs
    associated with a user's session. It tracks modifications to its contents and
    supports standard dictionary operations such as getting, setting, deleting, and
    updating items. The session data can be converted to a dictionary using `to_dict()`.

    Attributes:
        modified (bool): Indicates whether the session data has been modified.

    Methods:
        get(name, default=None): Retrieves a value by key, returning default if not
            found.
        set(name, value): Sets a value for a given key and marks the session as
            modified.
        update(values): Updates the session with multiple key-value pairs.
        clear(): Removes all items from the session and marks it as modified.
        to_dict(): Returns a shallow copy of the session data as a dictionary.
    """

    def __init__(self, values: Mapping[str, Any] | None = None) -> None:
        if values is None:
            values = {}
        self._modified = False
        self._values = dict(values)

    @property
    def modified(self) -> bool:
        return self._modified

    def get(self, name: str, default: Any = None) -> Any:
        return self._values.get(name, default)

    def set(self, name: str, value: Any) -> None:
        self._modified = True
        self._values[name] = value

    def update(self, values: Mapping[str, Any]) -> None:
        self._modified = True
        self._values.update(values)

    def __getitem__(self, name: str) -> Any:
        return self._values[name]

    def __setitem__(self, name: str, value: Any) -> None:
        self._modified = True
        self._values[name] = value

    def __delitem__(self, name: str) -> None:
        self._modified = True
        del self._values[name]

    def __contains__(self, name: str) -> bool:
        return name in self._values

    def __len__(self) -> int:
        return len(self._values)

    def __eq__(self, o: object) -> bool:
        if self is o:
            return True
        if isinstance(o, Session):
            return self._values == o._values
        return self._values == o

    def clear(self) -> None:
        self._modified = True
        self._values.clear()

    def to_dict(self) -> dict[str, Any]:
        return self._values.copy()


class SessionSerializer(ABC):
    """
    Abstract base class for session serialization and deserialization.

    Implementations of this class provide methods to convert Session objects
    to and from their string representations, enabling storage and retrieval
    of session data in various formats (e.g., JSON, base64, etc.).
    """

    @abstractmethod
    def read(self, value: str) -> Session:
        """Creates an instance of Session from a string representation."""

    @abstractmethod
    def write(self, session: Session) -> str:
        """Creates the string representation of a session."""


class SessionStore(ABC):
    """
    Abstract base class for session storage backends.

    Implementations of this class define how sessions are loaded from and saved to
    a storage medium (such as cookies, databases, or distributed caches) during the
    request-response cycle.
    """

    @abstractmethod
    async def load(self, request: Request) -> Session:
        """Load the session for the given request."""

    @abstractmethod
    async def save(
        self, request: Request, response: Response, session: Session
    ) -> None:
        """Save the session related to the given request-response cycle."""



import asyncio

from blacksheep.contents import Content
from blacksheep.server.application import Application
from blacksheep.server.responses import Response
from blacksheep.testing.simulator import AbstractTestSimulator, TestSimulator
from blacksheep.testing.websocket import TestWebSocket

from .helpers import CookiesType, HeadersType, QueryType


class TestClient:
    # Setting this dunder variable
    # We tell to pytest don't discover this up
    __test__ = False

    def __init__(
        self, app: Application, test_simulator: AbstractTestSimulator | None = None
    ):
        self._test_simulator = test_simulator or TestSimulator(app)

    async def get(
        self,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        cookies: CookiesType = None,
    ) -> Response:
        """Simulates HTTP GET method"""
        return await self._test_simulator.send_request(
            method="GET",
            path=path,
            headers=headers,
            query=query,
            cookies=cookies,
            content=None,
        )

    async def post(
        self,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        content: Content | None = None,
        cookies: CookiesType = None,
    ) -> Response:
        """Simulates HTTP POST method"""
        return await self._test_simulator.send_request(
            method="POST",
            path=path,
            headers=headers,
            query=query,
            cookies=cookies,
            content=content,
        )

    async def patch(
        self,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        content: Content | None = None,
        cookies: CookiesType = None,
    ) -> Response:
        """Simulates HTTP PATCH method"""
        return await self._test_simulator.send_request(
            method="PATCH",
            path=path,
            headers=headers,
            query=query,
            cookies=cookies,
            content=content,
        )

    async def put(
        self,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        content: Content | None = None,
        cookies: CookiesType = None,
    ) -> Response:
        """Simulates HTTP PUT method"""
        return await self._test_simulator.send_request(
            method="PUT",
            path=path,
            headers=headers,
            query=query,
            content=content,
            cookies=cookies,
        )

    async def delete(
        self,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        content: Content | None = None,
        cookies: CookiesType = None,
    ) -> Response:
        """Simulates HTTP DELETE method"""
        return await self._test_simulator.send_request(
            method="DELETE",
            path=path,
            headers=headers,
            query=query,
            content=content,
            cookies=cookies,
        )

    async def options(
        self,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        cookies: CookiesType = None,
    ) -> Response:
        """Simulates HTTP OPTIONS method"""
        return await self._test_simulator.send_request(
            method="OPTIONS",
            path=path,
            headers=headers,
            query=query,
            content=None,
            cookies=cookies,
        )

    async def head(
        self,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        cookies: CookiesType = None,
    ) -> Response:
        """Simulates HTTP HEAD method"""
        return await self._test_simulator.send_request(
            method="HEAD",
            path=path,
            headers=headers,
            query=query,
            content=None,
            cookies=cookies,
        )

    async def trace(
        self,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        cookies: CookiesType = None,
    ) -> Response:
        """Simulates HTTP TRACE method"""
        return await self._test_simulator.send_request(
            method="TRACE",
            path=path,
            headers=headers,
            query=query,
            content=None,
            cookies=cookies,
        )

    def websocket_connect(
        self,
        path: str,
        headers: HeadersType = None,
        query: QueryType = None,
        cookies: CookiesType = None,
    ) -> TestWebSocket:
        return self._test_simulator.websocket_connect(
            path=path,
            headers=headers,
            query=query,
            content=None,
            cookies=cookies,
        )

    async def websocket_all_closed(self):
        await asyncio.gather(*self._test_simulator.websocket_tasks)
        self._test_simulator.websocket_tasks = []



import base64

import pytest
from essentials.secrets import Secret
from guardpost import Identity

from blacksheep.server.authentication.basic import (
    BasicAuthentication,
    BasicCredentials,
    BasicCredentialsProvider,
)
from blacksheep.testing.helpers import get_example_scope
from blacksheep.testing.messages import MockReceive, MockSend
from tests.utils.application import FakeApplication


# Test fixtures
@pytest.fixture
def basic_credentials():
    return BasicCredentials(
        username="testuser",
        password=Secret("password123", direct_value=True),
        claims={"name": "Test User", "email": "test@example.com"},
        roles=["user", "reader"],
    )


@pytest.fixture
def admin_credentials():
    return BasicCredentials(
        username="admin",
        password=Secret("admin123", direct_value=True),
        claims={"name": "Admin User", "email": "admin@example.com"},
        roles=["admin", "user"],
    )


class MockCredentialsProvider(BasicCredentialsProvider):
    def __init__(self, credentials: list[BasicCredentials]):
        self._credentials = credentials

    async def get_credentials(self) -> list[BasicCredentials]:
        return self._credentials


class FailingCredentialsProvider(BasicCredentialsProvider):
    async def get_credentials(self) -> list[BasicCredentials]:
        raise Exception("Database connection failed")


# BasicCredentials tests
def test_basic_credentials_creation():
    credentials = BasicCredentials(
        username="testuser",
        password=Secret("password123", direct_value=True),
        claims={"custom": "claim"},
        roles=["user"],
    )

    assert credentials.username == "testuser"
    assert credentials.claims == {"custom": "claim"}
    assert credentials.roles == ["user"]


def test_basic_credentials_defaults():
    credentials = BasicCredentials(
        username="testuser",
        password=Secret("password123", direct_value=True),
    )

    assert credentials.username == "testuser"
    assert credentials.claims == {}
    assert credentials.roles == []


def test_basic_credentials_to_header_value():
    credentials = BasicCredentials(
        username="testuser",
        password=Secret("password123", direct_value=True),
    )

    header_value = credentials.to_header_value()
    expected = "Basic " + base64.b64encode(b"testuser:password123").decode("utf-8")

    assert header_value == expected


def test_basic_credentials_match_success():
    credentials = BasicCredentials(
        username="testuser",
        password=Secret("password123", direct_value=True),
    )

    assert credentials.match("testuser", "password123") is True


def test_basic_credentials_match_wrong_username():
    credentials = BasicCredentials(
        username="testuser",
        password=Secret("password123", direct_value=True),
    )

    assert credentials.match("wronguser", "password123") is False


def test_basic_credentials_match_wrong_password():
    credentials = BasicCredentials(
        username="testuser",
        password=Secret("password123", direct_value=True),
    )

    assert credentials.match("testuser", "wrongpassword") is False


# BasicAuthentication tests
def test_basic_authentication_creation_with_credentials():
    credentials = BasicCredentials(
        username="test",
        password=Secret("pass", direct_value=True),
    )

    auth = BasicAuthentication(credentials)
    assert auth.scheme == "Basic"
    assert len(auth._credentials) == 1


def test_basic_authentication_creation_with_provider():
    provider = MockCredentialsProvider([])
    auth = BasicAuthentication(credentials_provider=provider)
    assert auth._credentials_provider is provider


def test_basic_authentication_creation_with_custom_scheme():
    credentials = BasicCredentials(
        username="test",
        password=Secret("pass", direct_value=True),
    )

    auth = BasicAuthentication(credentials, scheme="CustomBasic")
    assert auth.scheme == "CustomBasic"


def test_basic_authentication_creation_requires_credentials_or_provider():
    with pytest.raises(
        ValueError, match="Either credentials or credentials_provider must be provided"
    ):
        BasicAuthentication()


async def test_basic_authentication_success(app: FakeApplication, basic_credentials):
    app.use_authentication().add(BasicAuthentication(basic_credentials))

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    # Create valid authorization header
    auth_header = basic_credentials.to_header_value().encode()

    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Authorization", auth_header)],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is True
    assert identity["sub"] == "testuser"
    assert identity["name"] == "Test User"
    assert identity["email"] == "test@example.com"
    assert identity["roles"] == ["user", "reader"]


async def test_basic_authentication_no_header(app: FakeApplication, basic_credentials):
    app.use_authentication().add(BasicAuthentication(basic_credentials))

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    await app(
        get_example_scope("GET", "/"),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is False


async def test_basic_authentication_wrong_scheme(
    app: FakeApplication, basic_credentials
):
    app.use_authentication().add(BasicAuthentication(basic_credentials))

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Authorization", b"Bearer token123")],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is False


async def test_basic_authentication_invalid_base64(
    app: FakeApplication, basic_credentials
):
    app.use_authentication().add(BasicAuthentication(basic_credentials))

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Authorization", b"Basic invalid-base64!")],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is False


async def test_basic_authentication_malformed_credentials(
    app: FakeApplication, basic_credentials
):
    app.use_authentication().add(BasicAuthentication(basic_credentials))

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    # Valid base64 but missing colon separator
    malformed_creds = base64.b64encode(b"usernamewithoutcolon").decode()

    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Authorization", f"Basic {malformed_creds}".encode())],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is False


async def test_basic_authentication_wrong_credentials(
    app: FakeApplication, basic_credentials
):
    app.use_authentication().add(BasicAuthentication(basic_credentials))

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    # Valid format but wrong credentials
    wrong_creds = base64.b64encode(b"wronguser:wrongpass").decode()

    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Authorization", f"Basic {wrong_creds}".encode())],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is False


async def test_basic_authentication_multiple_credentials(
    app: FakeApplication, basic_credentials, admin_credentials
):
    app.use_authentication().add(
        BasicAuthentication(basic_credentials, admin_credentials)
    )

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    # Test with first credentials
    auth_header = basic_credentials.to_header_value().encode()
    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Authorization", auth_header)],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity["sub"] == "testuser"

    # Test with second credentials
    auth_header = admin_credentials.to_header_value().encode()
    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Authorization", auth_header)],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity["sub"] == "admin"
    assert "admin" in identity["roles"]


async def test_basic_authentication_with_provider(
    app: FakeApplication, basic_credentials
):
    provider = MockCredentialsProvider([basic_credentials])
    app.use_authentication().add(BasicAuthentication(credentials_provider=provider))

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    auth_header = basic_credentials.to_header_value().encode()

    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Authorization", auth_header)],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is True
    assert identity["sub"] == "testuser"


async def test_basic_authentication_with_failing_provider(app: FakeApplication):
    provider = FailingCredentialsProvider()
    app.use_authentication().add(BasicAuthentication(credentials_provider=provider))

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    auth_header = base64.b64encode(b"user:pass").decode()

    # Must not eat the exception
    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Authorization", f"Basic {auth_header}".encode())],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 500


async def test_basic_authentication_empty_provider(app: FakeApplication):
    provider = MockCredentialsProvider([])
    app.use_authentication().add(BasicAuthentication(credentials_provider=provider))

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    auth_header = base64.b64encode(b"user:pass").decode()

    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Authorization", f"Basic {auth_header}".encode())],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is False


def test_basic_authentication_custom_description():
    credentials = BasicCredentials(
        username="test",
        password=Secret("pass", direct_value=True),
    )

    auth = BasicAuthentication(credentials, description="Custom Basic Auth")
    assert auth.description == "Custom Basic Auth"


async def test_basic_authentication_unicode_credentials(app: FakeApplication):
    # Test with unicode characters in username/password
    unicode_credentials = BasicCredentials(
        username="тест",
        password=Secret("пароль123", direct_value=True),
        claims={"name": "Unicode User"},
    )

    app.use_authentication().add(BasicAuthentication(unicode_credentials))

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    auth_header = unicode_credentials.to_header_value().encode()

    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Authorization", auth_header)],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is True
    assert identity["sub"] == "тест"
    assert identity["name"] == "Unicode User"


async def test_basic_authentication_password_with_colon(app: FakeApplication):
    # Test password containing colon character
    credentials = BasicCredentials(
        username="testuser",
        password=Secret("pass:word:123", direct_value=True),
    )

    app.use_authentication().add(BasicAuthentication(credentials))

    identity: Identity | None = None

    @app.router.get("/")
    async def home(request):
        nonlocal identity
        identity = request.user
        return None

    await app.start()

    auth_header = credentials.to_header_value().encode()

    await app(
        get_example_scope(
            "GET",
            "/",
            extra_headers=[(b"Authorization", auth_header)],
        ),
        MockReceive(),
        MockSend(),
    )

    assert app.response.status == 204
    assert identity is not None
    assert identity.is_authenticated() is True
    assert identity["sub"] == "testuser"



from typing import Optional

class InvalidOperation(Exception):
    def __init__(self, message: str, inner_exception: object = None):
        super().__init__(message)
        self.inner_exception = inner_exception

class HTTPException(Exception):
    def __init__(self, status: int, message: str = "HTTP Exception"):
        super().__init__(message)
        self.status = status

class BadRequest(HTTPException):
    def __init__(self, message: str):
        super().__init__(400, message)

class Unauthorized(HTTPException):
    def __init__(self, message: str = "Unauthorized"):
        super().__init__(401, message)

class Forbidden(HTTPException):
    def __init__(self, message: str = "Forbidden"):
        super().__init__(403, message)

class Conflict(HTTPException):
    def __init__(self, message: str = "Conflict"):
        super().__init__(409, message)

class BadRequestFormat(BadRequest):
    def __init__(self, message: str, inner_exception: object = None):
        super().__init__(message)
        self.inner_exception = inner_exception

class RangeNotSatisfiable(HTTPException):
    def __init__(self, message: str = "Range Not Satisfiable"):
        super().__init__(416, message)

class NotFound(HTTPException):
    def __init__(self):
        super().__init__(404)

class InvalidArgument(Exception):
    def __init__(self, message: str):
        super().__init__(message)

class MessageAborted(Exception):
    def __init__(self):
        super().__init__(
            "The message was aborted before the client sent its whole content."
        )

class InternalServerError(HTTPException):
    def __init__(
        self,
        message: str = "Internal Server Error",
        source_error: Exception | None = None,
    ):
        super().__init__(500, message)
        self.source_error = source_error

class NotImplementedByServer(HTTPException):
    def __init__(self, message: str = "Not Implemented"):
        super().__init__(501, message)

class FailedRequestError(HTTPException):
    def __init__(self, status: int, data: str) -> None:
        super().__init__(
            status,
            f"The response status code does not indicate success: {status}. Response body: {data}",
        )
        self.data = data

class InvalidExceptionHandler(Exception):
    pass



"""
This module provides a strategy and classes to convert client input into expected types.
This is used to offer a good user experience to developers: rather than parsing raw
input values from requests manually, they can declare the expected types using Python
type annotations, and the code tries to obtain objects of the expected type.

The following code only offers default implementations that should work in most cases,
the user can still define custom logic to parse input values.
"""

import inspect
from abc import ABC, abstractmethod
from collections.abc import Sequence as SequenceABC
from dataclasses import fields, is_dataclass
from datetime import date, datetime
from functools import lru_cache
from typing import (
    Any,
    Callable,
    List,
    Literal,
    Sequence,
    get_args,
    get_origin,
    get_type_hints,
)
from urllib.parse import unquote
from uuid import UUID

from blacksheep.exceptions import BadRequest
from blacksheep.server.bindings.dates import parse_datetime
from blacksheep.utils import ensure_str

try:
    # Supported only in Python >= 3.11
    from enum import IntEnum, StrEnum
except ImportError:
    IntEnum = None
    StrEnum = None

try:
    # Pydantic v2 support
    from pydantic import BaseModel
except ImportError:
    BaseModel = None


class TypeConverter(ABC):
    """
    Base class for types that converts string reprensentations of values into
    instances of specific types.
    """

    @abstractmethod
    def can_convert(self, expected_type) -> bool:
        """Returns True if this converter can handle the expected type."""

    @abstractmethod
    def convert(self, value, expected_type) -> Any:
        """Converts a str value into an object of the expected type."""


class StrConverter(TypeConverter):

    def can_convert(self, expected_type) -> bool:
        return expected_type is str or str(expected_type) == "~T"

    def convert(self, value, expected_type) -> Any:
        if value is None:
            return None
        if "%" in value:
            return unquote(value)
        return value


class InitTypeConverter(TypeConverter):

    def __init__(self, obj_type) -> None:
        self._handled_type = obj_type

    def can_convert(self, expected_type) -> bool:
        return expected_type is self._handled_type

    def convert(self, value, expected_type) -> Any:
        return self._handled_type(value) if value is not None else None


class BoolConverter(TypeConverter):

    def can_convert(self, expected_type) -> bool:
        return expected_type is bool

    def convert(self, value, expected_type) -> Any:
        if value is None:
            return None
        if not isinstance(value, str):
            return bool(value)
        if value.lower() in ("true", "1"):
            return True
        elif value.lower() in ("false", "0"):
            return False
        else:
            raise ValueError(f"Cannot convert {value!r} to {expected_type.__name__}")


class IntConverter(InitTypeConverter):

    def __init__(self) -> None:
        super().__init__(int)


class FloatConverter(InitTypeConverter):

    def __init__(self) -> None:
        super().__init__(float)


class UUIDConverter(InitTypeConverter):

    def __init__(self) -> None:
        super().__init__(UUID)


class BytesConverter(TypeConverter):

    def __init__(self, encoding: str = "utf8") -> None:
        self._encoding = encoding

    def can_convert(self, expected_type) -> bool:
        return expected_type is bytes

    def convert(self, value, expected_type) -> Any:
        return value.encode(self._encoding) if value else None


class DateTimeConverter(TypeConverter):

    def __init__(self, parser_fn: Callable[[str], datetime] = parse_datetime) -> None:
        self.parser_fn = parser_fn

    def can_convert(self, expected_type) -> bool:
        return expected_type is datetime

    def convert(self, value, expected_type) -> Any:
        return self.parser_fn(value) if value else None


class DateConverter(TypeConverter):

    def __init__(self, parser_fn: Callable[[str], datetime] = parse_datetime) -> None:
        self.parser_fn = parser_fn

    def can_convert(self, expected_type) -> bool:
        return expected_type is date

    def convert(self, value, expected_type) -> Any:
        return self.parser_fn(value).date() if value else None


class EnumConverter(TypeConverter):

    @abstractmethod
    def parse_by_value(self, value, expected_type) -> Any:
        """Obtains an enum of the expected type by value."""

    @abstractmethod
    def parse_by_key(self, value, expected_type) -> Any:
        """Obtains an enum of the expected type by key."""

    def convert(self, value, expected_type) -> Any:
        if value is None:
            return None
        try:
            return self.parse_by_value(value, expected_type)
        except ValueError:
            try:
                return self.parse_by_key(value, expected_type)
            except KeyError:
                raise BadRequest(f"{value} is not a valid {expected_type.__name__}")


class StrEnumConverter(EnumConverter):

    def can_convert(self, expected_type) -> bool:
        return isinstance(expected_type, type) and issubclass(expected_type, StrEnum)

    def parse_by_value(self, value: str, expected_type) -> Any:
        return expected_type(value)

    def parse_by_key(self, value: str, expected_type) -> Any:
        return expected_type[value]


class IntEnumConverter(EnumConverter):

    def can_convert(self, expected_type) -> bool:
        return isinstance(expected_type, type) and issubclass(expected_type, IntEnum)

    def parse_by_value(self, value: str, expected_type) -> Any:
        return expected_type(int(ensure_str(value)))

    def parse_by_key(self, value: str, expected_type) -> Any:
        return expected_type[value]


class LiteralConverter(TypeConverter):

    def __init__(self, case_insensitive: bool = False) -> None:
        self.case_insensitive = case_insensitive

    def can_convert(self, expected_type) -> bool:
        return (
            hasattr(expected_type, "__origin__") and expected_type.__origin__ is Literal
        )

    def convert(self, value, expected_type) -> Any:
        if value is None:
            return None
        allowed = get_args(expected_type)
        for allowed_value in allowed:
            if str(allowed_value) == value:
                return allowed_value
            if (
                self.case_insensitive
                and isinstance(allowed_value, str)
                and allowed_value.lower() == value.lower()
            ):
                return allowed_value
        raise ValueError(f"{value!r} is not a valid {expected_type}")


class _AsIsConverter(TypeConverter):
    def can_convert(self, expected_type) -> bool:
        return True

    def convert(self, value, expected_type) -> Any:
        return value


_as_is_converter = _AsIsConverter()


class DictConverter(TypeConverter):
    """
    Converter for dict[K, V] where both K and V can be handled by converters.
    Supports string keys and hashable class keys.
    """

    def can_convert(self, expected_type) -> bool:
        return _get_origin(expected_type) is dict

    def convert(self, value, expected_type) -> Any:
        if value is None:
            return None

        key_type, value_type = _get_args(expected_type)

        # Find the appropriate converters
        key_converter = get_converter(key_type)
        value_converter = get_converter(value_type)

        return {
            key_converter.convert(k, key_type): value_converter.convert(v, value_type)
            for k, v in value.items()
        }


_dict_converter = DictConverter()


@lru_cache(maxsize=None)
def _get_signature(cls):
    return inspect.signature(cls.__init__)


@lru_cache(maxsize=None)
def _get_type_hints(cls):
    return get_type_hints(cls.__init__)


@lru_cache(maxsize=None)
def _get_dataclass_fields(cls):
    return fields(cls)


@lru_cache(maxsize=None)
def _is_pydantic_model(cls):
    return BaseModel is not None and inspect.isclass(cls) and issubclass(cls, BaseModel)


@lru_cache(maxsize=None)
def _get_args(cls):
    return get_args(cls)


@lru_cache(maxsize=None)
def _get_origin(cls):
    return get_origin(cls)


@lru_cache(maxsize=None)
def _is_dataclass(cls):
    return is_dataclass(cls)


class ClassConverter(TypeConverter):
    """
    Converts dictionaries to instances of desired types, supporting Pydantic models,
    Python dataclasses or plain user-defined classes.

    This converter handles common scenarios for basic type conversion from dictionaries
    to class instances. It supports:
    - Dataclasses with simple field types
    - Plain classes with __init__ parameters
    - Nested classes and dataclasses

    Important limitations:
    - This converter is NOT designed to support all possible type conversion scenarios
    - It handles only straightforward cases with basic type annotations
    - Complex type validation, nested generics, and advanced typing constructs
      are not fully supported

    For complex type conversion scenarios, use:
    - Pydantic models (recommended): Provides comprehensive validation,
      advanced typing support, and better error messages
    - Custom type converters: Define explicit conversion logic for your specific needs
    - Explicit conversion in your input types classes

    This converter ignores extra fields in the input dictionary that don't match
    class parameters or dataclass fields.
    """

    def _from_dict(self, cls, data: dict | list):
        """Convert dict to plain class or dataclass, ignoring extra fields"""

        if _is_pydantic_model(cls):
            return cls.model_validate(data)

        # here it is sufficient to handle list because input from client can only
        # be parsed as list in most cases (like after parsing JSON or XML), not other
        # types of sequences like tuple
        if isinstance(data, list):
            # require a type hint to work
            obj_type_hint = _get_args(cls)
            if obj_type_hint:
                # good…
                obj_type_converter = get_converter(obj_type_hint[0])
                return [
                    obj_type_converter.convert(datum, obj_type_hint[0])
                    for datum in data
                ]
            else:
                # return data as-is (let if fail downstream if it must)
                return data

        if _dict_converter.can_convert(cls):
            return _dict_converter.convert(data, cls)

        if not isinstance(data, dict):
            return data

        # Handle dataclasses
        if _is_dataclass(cls):
            return self._handle_dataclass(cls, data)

        # Handle plain classes
        return self._handle_plain_class(cls, data)

    def _handle_dataclass(self, cls, data):
        field_values = {}
        for field in _get_dataclass_fields(cls):
            if field.name in data:
                field_type = field.type
                value = data[field.name]
                if value is None:
                    field_values[field.name] = None
                else:
                    converter = get_converter(field_type)
                    field_values[field.name] = converter.convert(value, field_type)
        return cls(**field_values)

    def _handle_plain_class(self, cls, data):
        # Get type hints from __init__
        sig = _get_signature(cls)
        type_hints = _get_type_hints(cls)

        init_params = {}
        for param_name, _ in sig.parameters.items():
            if param_name == "self":
                continue

            if param_name in data:
                value = data[param_name]

                # Check if parameter has a type hint that's a class
                if param_name in type_hints:
                    param_type = type_hints[param_name]
                    converter = get_converter(param_type)
                    init_params[param_name] = converter.convert(value, param_type)
                else:
                    init_params[param_name] = value
        return cls(**init_params)

    @lru_cache(maxsize=None)
    def can_convert(self, expected_type) -> bool:
        # Must be a class
        if not inspect.isclass(expected_type):
            return False

        if is_dataclass(expected_type) or _is_pydantic_model(expected_type):
            return True

        # Exclude built-in types
        if expected_type.__module__ == "builtins":
            return False

        # Exclude typing module generics
        if get_origin(expected_type) is not None:
            return False

        # Must have a callable __init__ method
        if not hasattr(expected_type, "__init__"):
            return False

        return True

    def convert(self, value, expected_type) -> Any:
        if value is None:
            return None

        return self._from_dict(expected_type, value)


class ListConverter(TypeConverter):
    """
    Converter for list[T], Sequence[T], and tuple[T] where T
    is handled by class_converters (DataClass, Pydantic models, plain classes).
    """

    def __init__(self, supported_origins=None):
        if supported_origins is None:
            supported_origins = {list, List, Sequence, SequenceABC, tuple}
        self.supported_origins = frozenset(supported_origins)

    def can_convert(self, expected_type) -> bool:
        origin = _get_origin(expected_type)
        if origin not in self.supported_origins:
            return False

        # Get the item type
        args = _get_args(expected_type)
        if not args:
            return False

        return True

    def convert(self, value, expected_type) -> Any:
        if value is None:
            return None

        origin = _get_origin(expected_type)
        item_type = _get_args(expected_type)[0]

        item_converter = get_converter(item_type)

        converted_items = [item_converter.convert(item, item_type) for item in value]

        if origin in {list, List}:
            return converted_items
        elif origin is tuple:
            return tuple(converted_items)
        else:
            return converted_items


converters: list[TypeConverter] = [
    BoolConverter(),
    BytesConverter(),
    DateConverter(),
    DateTimeConverter(),
    IntConverter(),
    FloatConverter(),
    LiteralConverter(),
    StrConverter(),
    UUIDConverter(),
]


if StrEnum is not None and IntEnum is not None:
    # Python > 3.10
    converters.append(IntEnumConverter())
    converters.append(StrEnumConverter())


class_converters: list[TypeConverter] = [
    ClassConverter(),
    ListConverter(),
    DictConverter(),
]


@lru_cache(maxsize=None)
def get_converter(cls) -> TypeConverter:
    _all_converters = converters + class_converters

    for converter in _all_converters:
        if converter.can_convert(cls):
            return converter
    return _as_is_converter



from __future__ import annotations

import asyncio
from typing import Any, AnyStr, MutableMapping

from blacksheep.settings.json import json_settings


class TestWebSocket:
    def __init__(self):
        self.send_queue = asyncio.Queue()
        self.receive_queue = asyncio.Queue()

    async def _send(self, data: MutableMapping[str, Any]) -> None:
        await self.send_queue.put(data)

    async def _receive(self) -> MutableMapping[str, AnyStr]:
        return await self.receive_queue.get()

    async def send(self, data: MutableMapping[str, Any]) -> None:
        await self.receive_queue.put(data)

    async def send_text(self, data: str) -> None:
        await self.send({"type": "websocket.send", "text": data})

    async def send_bytes(self, data: bytes) -> None:
        await self.send({"type": "websocket.send", "bytes": data})

    async def send_json(self, data: MutableMapping[Any, Any]) -> None:
        await self.send(
            {
                "type": "websocket.send",
                "text": json_settings.dumps(data),
            }
        )

    async def receive(self) -> MutableMapping[str, AnyStr]:
        return await self.send_queue.get()

    async def receive_text(self) -> str:
        data = await self.receive()
        assert data["type"] == "websocket.send"
        return data["text"]

    async def receive_bytes(self) -> bytes:
        data = await self.receive()
        assert data["type"] == "websocket.send"
        return data["bytes"]

    async def receive_json(self) -> MutableMapping[str, Any]:
        data = await self.receive()
        assert data["type"] == "websocket.send"
        return json_settings.loads(data["text"])

    async def __aenter__(self) -> TestWebSocket:
        await self.send({"type": "websocket.connect"})
        received = await self.receive()
        assert received.get("type") == "websocket.accept"
        return self

    async def __aexit__(self, exc_type, exc_value, exc_tb) -> None:
        await self.send(
            {
                "type": "websocket.disconnect",
                "code": 1000,
                "reason": "TestWebSocket context closed",
            },
        )



from datetime import datetime


def parse_datetime(value: str) -> datetime:
    """
    This function parse ISO 8601 datetime strings.
    It only supports the most common ISO formats:

    - "%Y-%m-%dT%H:%M:%S.%f"
    - "%Y-%m-%dT%H:%M:%S"
    - "%Y-%m-%d"
    """
    value_len = len(value)
    # The following functions, unlike fromisoformat, accept parts without leading 0.
    # Try parsing with microseconds (minimum length: 16, like in "2025-6-9T8:6:3.1")
    if value_len >= 16:
        try:
            return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f")
        except ValueError:
            pass

    # Try parsing without microseconds (minimum length: 14, like in "2025-6-9T8:6:3")
    if value_len >= 14:
        try:
            return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S")
        except ValueError:
            pass

    # Try parsing date only (minimum length: 8, like in "2025-6-9")
    if value_len >= 8:
        try:
            return datetime.strptime(value, "%Y-%m-%d")
        except ValueError:
            pass

    raise ValueError("Could not parse datetime string.")



from blacksheep.settings.json import json_settings

from .abc import Session, SessionSerializer


class JSONSerializer(SessionSerializer):
    """
    Serializes and deserializes Session objects using JSON format.

    This serializer uses the application's JSON settings to convert session data
    to and from JSON strings, enabling storage and retrieval of session information
    in a widely supported text format.
    """

    def read(self, value: str) -> Session:
        return Session(json_settings.loads(value))

    def write(self, session: Session) -> str:
        return json_settings.dumps(session.to_dict())



import sys
from dataclasses import dataclass
from inspect import Parameter, _ParameterKind
from typing import AsyncIterable, Sequence

import pytest
from guardpost import Identity, User
from pytest import raises
from rodi import Container, Services, inject

from blacksheep import Request
from blacksheep.server.bindings import (
    Binder,
    BoundValue,
    ExactBinder,
    FromHeader,
    FromJSON,
    FromQuery,
    FromRoute,
    FromServices,
    HeaderBinder,
    IdentityBinder,
    JSONBinder,
    QueryBinder,
    RequestBinder,
    RouteBinder,
    ServiceBinder,
)
from blacksheep.server.normalization import (
    AmbiguousMethodSignatureError,
    NormalizationError,
    RouteBinderMismatch,
    UnsupportedSignatureError,
    _check_union,
    _get_raw_bound_value_type,
    get_asyncgen_yield_type,
    get_binders,
    normalize_handler,
    normalize_middleware,
)
from blacksheep.server.routing import Route


class Pet:
    def __init__(self, name):
        self.name = name


class Cat(Pet): ...


class Dog(Pet): ...


@dataclass
class SomeService:
    pass


def test_parameters_get_binders_default_query():
    def handler(a, b, c): ...

    binders = get_binders(Route(b"/", handler), Container())

    assert all(isinstance(binder, QueryBinder) for binder in binders)
    assert binders[0].parameter_name == "a"
    assert binders[1].parameter_name == "b"
    assert binders[2].parameter_name == "c"


@pytest.mark.parametrize(
    "annotation_type", [Identity, User, User | None, Identity | None]
)
def test_identity_binder_by_param_type(annotation_type):
    async def handler(param): ...

    handler.__annotations__["param"] = annotation_type

    binders = get_binders(Route(b"/", handler), Container())

    assert isinstance(binders[0], IdentityBinder)


def test_parameters_get_binders_from_route():
    def handler(a, b, c): ...

    binders = get_binders(Route(b"/:a/:b/:c", handler), Container())

    assert all(isinstance(binder, RouteBinder) for binder in binders)
    assert binders[0].parameter_name == "a"
    assert binders[1].parameter_name == "b"
    assert binders[2].parameter_name == "c"


def test_parameters_get_binders_from_services_by_name():
    def handler(a, b, c): ...

    binders = get_binders(
        Route(b"/", handler), {"a": object(), "b": object(), "c": object()}
    )

    assert all(isinstance(binder, ServiceBinder) for binder in binders)
    assert binders[0].expected_type == "a"
    assert binders[1].expected_type == "b"
    assert binders[2].expected_type == "c"


def test_parameters_get_binders_from_services_by_type():
    def handler(a: str, b: int, c: Cat): ...

    binders = get_binders(
        Route(b"/", handler), {str: object(), int: object(), Cat: object()}
    )

    assert all(isinstance(binder, ServiceBinder) for binder in binders)
    assert binders[0].expected_type is str
    assert binders[1].expected_type is int
    assert binders[2].expected_type is Cat


def test_parameters_get_binders_from_body():
    def handler(a: Cat): ...

    binders = get_binders(Route("/", handler), Container())
    assert len(binders) == 1
    binder = binders[0]

    assert isinstance(binder, JSONBinder)
    assert binder.expected_type is Cat
    assert binder.required is True


def test_parameters_get_binders_from_body_optional():
    def handler(a: Cat | None): ...

    binders = get_binders(Route(b"/", handler), Container())
    assert len(binders) == 1
    binder = binders[0]

    assert isinstance(binder, JSONBinder)
    assert binder.expected_type is Cat
    assert binder.required is False


def test_parameters_get_binders_simple_types_default_from_query():
    def handler(a: str, b: int, c: bool): ...

    binders = get_binders(Route(b"/", handler), Container())

    assert all(isinstance(binder, QueryBinder) for binder in binders)
    assert binders[0].parameter_name == "a"
    assert binders[0].expected_type == str
    assert binders[1].parameter_name == "b"
    assert binders[1].expected_type == int
    assert binders[2].parameter_name == "c"
    assert binders[2].expected_type == bool


def test_parameters_get_binders_list_types_default_from_query():
    def handler(a: list[str], b: list[int], c: list[bool]): ...

    binders = get_binders(Route(b"/", handler), Container())

    assert all(isinstance(binder, QueryBinder) for binder in binders)
    assert binders[0].parameter_name == "a"
    assert binders[0].expected_type == list[str]
    assert binders[1].parameter_name == "b"
    assert binders[1].expected_type == list[int]
    assert binders[2].parameter_name == "c"
    assert binders[2].expected_type == list[bool]


def test_parameters_get_binders_list_types_default_from_query_optional():
    def handler(a: list[str] | None): ...

    binders = get_binders(Route(b"/", handler), Container())

    assert all(isinstance(binder, QueryBinder) for binder in binders)
    assert all(binder.required is False for binder in binders)
    assert binders[0].parameter_name == "a"
    assert binders[0].expected_type == list[str]


def test_parameters_get_binders_list_types_default_from_query_required():
    def handler(a: list[str]): ...

    binders = get_binders(Route(b"/", handler), Container())

    assert all(isinstance(binder, QueryBinder) for binder in binders)
    assert all(binder.required for binder in binders)


def test_parameters_get_binders_sequence_types_default_from_query():
    def handler(a: Sequence[str], b: Sequence[int], c: Sequence[bool]): ...

    binders = get_binders(Route(b"/", handler), Container())

    assert all(isinstance(binder, QueryBinder) for binder in binders)
    assert binders[0].parameter_name == "a"
    assert binders[0].expected_type == Sequence[str]
    assert binders[1].parameter_name == "b"
    assert binders[1].expected_type == Sequence[int]
    assert binders[2].parameter_name == "c"
    assert binders[2].expected_type == Sequence[bool]


def test_throw_for_ambiguous_binder_multiple_from_body():
    def handler(a: Cat, b: Dog): ...

    with pytest.raises(AmbiguousMethodSignatureError):
        get_binders(Route(b"/", handler), Container())


def test_does_not_throw_for_forward_ref():
    def handler(a: "Cat"): ...

    get_binders(Route(b"/", handler), Container())

    def handler(a: list["str"]): ...

    get_binders(Route(b"/", handler), Container())

    def handler(a: list["Cat | None"]): ...

    get_binders(Route(b"/", handler), Container())

    def handler(a: FromQuery["str"]): ...

    get_binders(Route(b"/", handler), Container())


def test_combination_of_sources():
    def handler(
        a: FromQuery[list[str]],
        b: FromServices[Dog],
        c: FromJSON[Cat],
        d: FromRoute[str],
        e: FromHeader[str],
    ): ...

    container = Container()
    container.register(Dog, instance=Dog("Snoopy"))
    binders = get_binders(Route(b"/:d", handler), container)

    assert isinstance(binders[0], QueryBinder)
    assert isinstance(binders[1], ServiceBinder)
    assert isinstance(binders[2], JSONBinder)
    assert isinstance(binders[3], RouteBinder)
    assert isinstance(binders[4], HeaderBinder)
    assert binders[0].parameter_name == "a"
    assert binders[1].parameter_name == "b"
    assert binders[2].parameter_name == "c"
    assert binders[3].parameter_name == "d"
    assert binders[4].parameter_name == "e"


def test_implicit_from_services_only_when_annotation_is_none():
    def handler(dog): ...

    container = Container()
    container.register("dog", instance=Dog("Snoopy"))

    binders = get_binders(Route(b"/", handler), container)

    assert isinstance(binders[0], ServiceBinder)

    def handler(dog: str): ...

    binders = get_binders(Route(b"/", handler), container)

    assert isinstance(binders[0], QueryBinder)


def test_from_query_specific_name():
    class FromExampleQuery(FromQuery[str]):
        name = "example"

    @inject()
    def handler(a: FromExampleQuery): ...

    binders = get_binders(Route(b"/", handler), Container())
    binder = binders[0]

    assert isinstance(binder, QueryBinder)
    assert binder.expected_type is str
    assert binder.required is True
    assert binder.parameter_name == "example"


def test_from_query_unspecified_type():
    def handler(a: FromQuery): ...

    binders = get_binders(Route(b"/", handler), Container())
    binder = binders[0]

    assert isinstance(binder, QueryBinder)
    assert binder.expected_type == list[str]
    assert binder.required is True
    assert binder.parameter_name == "a"


def test_from_query_optional_type():
    def handler(a: FromQuery[str | None]): ...

    binders = get_binders(Route(b"/", handler), Container())
    binder = binders[0]

    assert isinstance(binder, QueryBinder)
    assert binder.expected_type is str
    assert binder.required is False
    assert binder.parameter_name == "a"


def test_from_query_optional_type_with_union():
    def handler(a: FromQuery[str | None]): ...

    binders = get_binders(Route(b"/", handler), Container())
    binder = binders[0]

    assert isinstance(binder, QueryBinder)
    assert binder.expected_type is str
    assert binder.required is False
    assert binder.parameter_name == "a"


def test_check_union():
    optional, value = _check_union(
        Parameter("foo", kind=_ParameterKind.POSITIONAL_ONLY),
        str | None,
        len,
    )
    assert optional is True
    assert value is str


@pytest.mark.skipif(sys.version_info < (3, 10), reason="requires python3.10 or higher")
def test_check_union_or_none():
    optional, value = _check_union(
        Parameter("foo", kind=_ParameterKind.POSITIONAL_ONLY),
        str | None,
        len,
    )
    assert optional is True
    assert value is str


def test_from_query_optional_list_type():
    def handler(a: FromQuery[list[str] | None]): ...

    binders = get_binders(Route(b"/", handler), Container())
    binder = binders[0]

    assert isinstance(binder, QueryBinder)
    assert binder.expected_type == list[str]
    assert binder.required is False
    assert binder.parameter_name == "a"


def test_from_header_specific_name():
    class FromExampleHeader(FromHeader[str]):
        name = "example"

    @inject()
    def handler(a: FromExampleHeader): ...

    binders = get_binders(Route(b"/", handler), Container())

    assert isinstance(binders[0], HeaderBinder)
    assert binders[0].expected_type is str
    assert binders[0].parameter_name == "example"


def test_from_header_accept_language_example():
    class AcceptLanguageHeader(FromHeader[str]):
        name = "accept-language"

    @inject()
    def handler(a: AcceptLanguageHeader): ...

    binders = get_binders(Route(b"/", handler), Container())

    assert isinstance(binders[0], HeaderBinder)
    assert binders[0].expected_type is str
    assert binders[0].parameter_name == "accept-language"


def test_raises_for_route_mismatch():
    def handler(a: FromRoute[str]): ...

    with raises(RouteBinderMismatch):
        get_binders(Route(b"/", handler), Container())


def test_raises_for_route_mismatch_2():
    def handler(a: FromRoute[str]): ...

    with raises(RouteBinderMismatch):
        get_binders(Route(b"/:b", handler), Container())


def test_raises_for_unsupported_union():
    def handler(a: FromRoute[str | int]): ...

    with raises(NormalizationError):
        get_binders(Route(b"/:b", handler), Container())


def test_request_binding():
    def handler(request): ...

    binders = get_binders(Route(b"/", handler), Container())

    assert isinstance(binders[0], RequestBinder)


def test_services_binding():
    app_services = Container()

    def handler(services):
        assert services is app_services

    binders = get_binders(Route(b"/", handler), app_services)

    assert isinstance(binders[0], ExactBinder)


async def test_services_from_normalization():
    app_services = Container()

    def handler(services):
        assert services is app_services
        return None

    method = normalize_handler(Route(b"/", handler), app_services)
    await method(None)


def test_raises_for_unsupported_signature():
    app_services = Container()

    def handler(services, *args):
        assert services is app_services
        return services

    def handler2(services, **kwargs):
        assert services is app_services
        return services

    def handler3(services, *, key_only):
        assert services is app_services
        return services

    with pytest.raises(UnsupportedSignatureError):
        normalize_handler(Route(b"/", handler), Container())

    with pytest.raises(UnsupportedSignatureError):
        normalize_handler(Route(b"/", handler2), Container())

    with pytest.raises(UnsupportedSignatureError):
        normalize_handler(Route(b"/", handler3), Container())


async def fake_handler(_):
    return "fake-handler-result"


async def test_middleware_normalization():
    container = Container()
    container.add_singleton(object)
    container.add_alias("context", object)

    fake_request = Request("GET", b"/", None)

    async def middleware(request, handler, context: object):
        assert request is fake_request
        assert handler is fake_handler
        assert context is container.resolve("context")
        return await handler(request)

    normalized = normalize_middleware(middleware, container)

    # NB: middlewares base signature is (request, handler)
    result = await normalized(fake_request, fake_handler)  # type: ignore
    assert result == "fake-handler-result"


async def test_middleware_normalization_2():
    services = Container()  # {"context": object(), "foo": object()}
    fake_request = Request("GET", b"/", None)

    async def middleware(context, foo):
        pass

    normalized = normalize_middleware(middleware, services)  # type: ignore

    # NB: middlewares base signature is (request, handler)
    result = await normalized(fake_request, fake_handler)  # type: ignore
    assert result == "fake-handler-result"


async def test_middleware_normalization_raises_for_sync_function():
    def faulty_middleware(request, handler):
        pass

    with pytest.raises(ValueError):
        normalize_middleware(faulty_middleware, Services())  # type: ignore


async def test_middleware_query_normalization():
    container = Container()
    container.register("context", instance=object())

    fake_request = Request("GET", b"/?example=Lorem", None)

    async def middleware(request, handler, example: FromQuery[str]):
        assert request is fake_request
        assert handler is fake_handler
        assert example.value == "Lorem"
        return await handler(request)

    normalized = normalize_middleware(middleware, container)

    # NB: middlewares base signature is (request, handler)
    result = await normalized(fake_request, fake_handler)  # type: ignore
    assert result == "fake-handler-result"


async def test_middleware_header_normalization():
    container = Container()
    container.register("context", instance=object())
    fake_request = Request("GET", b"/", [(b"example", b"Lorem")])

    async def middleware(request, handler, example: FromHeader[str]):
        assert request is fake_request
        assert handler is fake_handler
        assert example.value == "Lorem"
        return await handler(request)

    normalized = normalize_middleware(middleware, container)

    # NB: middlewares base signature is (request, handler)
    result = await normalized(fake_request, fake_handler)  # type: ignore
    assert result == "fake-handler-result"


async def test_middleware_normalization_no_parameters():
    container = Container()
    container.register("context", instance=object())
    fake_request = Request("GET", b"/", [(b"example", b"Lorem")])

    called = False

    async def middleware():
        nonlocal called
        called = True

    normalized = normalize_middleware(middleware, container)  # type: ignore

    # NB: middlewares base signature is (request, handler)
    # since our middleware above does not handle the next request handler,
    # it is called by the normalized method
    result = await normalized(fake_request, fake_handler)  # type: ignore
    assert called
    assert result == "fake-handler-result"


def test_middleware_not_normalized_if_signature_matches_expected_signature():
    async def middleware(request, handler):
        return await handler(request)

    normalized = normalize_middleware(middleware, Container())
    assert normalized is middleware


def test_get_raw_bound_value_type_fallsback_to_str():
    class Foo: ...

    assert _get_raw_bound_value_type(Foo) is str  # type: ignore


def test_normalization_with_service_json_route_param():
    def handler(
        foo_id: str,
        data: FromJSON[Cat],
        some_service: SomeService,
    ): ...

    container = Container()
    container.add_transient(SomeService)

    binders = get_binders(Route("/{foo_id}", handler), container)
    assert len(binders) == 3

    assert isinstance(binders[0], RouteBinder)
    assert binders[0].expected_type is str

    assert isinstance(binders[1], JSONBinder)
    assert binders[1].expected_type is Cat

    assert isinstance(binders[2], ServiceBinder)
    assert binders[2].expected_type is SomeService


def test_camel_case_route_parameter():
    def handler(statusKey: FromRoute[str]): ...

    binders = get_binders(Route(b"/:statusKey", handler), Container())

    assert isinstance(binders[0], RouteBinder)
    assert binders[0].parameter_name == "statusKey"


def test_pascal_case_route_parameter():
    def handler(StatusKey: FromRoute[str]): ...

    binders = get_binders(Route(b"/:StatusKey", handler), Container())

    assert isinstance(binders[0], RouteBinder)
    assert binders[0].parameter_name == "StatusKey"


async def test_binder_through_di():
    """
    Tests support for dependency injection when resolving custom binders.
    """

    class Foo: ...

    class FromExample(BoundValue[str]):
        pass

    class DIBinder(Binder):
        handle = FromExample

        def __init__(self, foo: Foo):
            super().__init__(str)
            self.foo = foo

        async def get_value(self, request: Request) -> str | None:
            assert isinstance(self.foo, Foo)
            return "Foo!"

    def handler(example: FromExample):
        assert example.value == "Foo!"
        return example.value

    container = Container()
    container.register(Foo)
    container.register(DIBinder)

    binders = get_binders(Route("/", handler), container)
    assert len(binders) == 1

    assert isinstance(binders[0], DIBinder)

    fake_request = Request("GET", b"/", None)

    method = normalize_handler(Route(b"/", handler), container)
    response = await method(fake_request)
    assert response is not None
    text = await response.text()
    assert text == "Foo!"


def test_get_asyncgen_yield_type():
    async def example_1(data: str) -> AsyncIterable[int]:
        for i in range(10):
            yield i

    async def example_2() -> AsyncIterable[str]:
        for i in range(10):
            yield str(i)

    assert get_asyncgen_yield_type(example_1) is int
    assert get_asyncgen_yield_type(example_2) is str



"""
This module provides classes to handle JWT Bearer authentication.
"""

import warnings
from typing import Sequence

from essentials.secrets import Secret
from guardpost import AuthenticationHandler, Identity, InvalidCredentialsError
from guardpost.jwks import KeysProvider
from guardpost.jwts import (
    AsymmetricJWTValidator,
    BaseJWTValidator,
    ExpiredAccessToken,
    InvalidAccessToken,
    SymmetricJWTValidator,
)
from jwt.exceptions import InvalidTokenError

from blacksheep.baseapp import get_logger
from blacksheep.messages import Request


class JWTBearerAuthentication(AuthenticationHandler):
    """
    AuthenticationHandler that can parse and verify JWT Bearer access tokens to identify
    users.

    JWTs are validated using either public RSA keys (asymmetric) or a shared secret
    (symmetric), but not both in the same instance. Keys can be fetched automatically
    from OpenID Connect (OIDC) discovery, if an `authority` is provided.

    Use separate instances of this class to support different authentication methods
    or identity providers.
    """

    def __init__(
        self,
        *,
        valid_audiences: Sequence[str],
        valid_issuers: Sequence[str] | None = None,
        authority: str | None = None,
        algorithms: Sequence[str] | None = None,
        require_kid: bool = True,
        keys_provider: KeysProvider | None = None,
        keys_url: str | None = None,
        cache_time: float = 10800,
        auth_mode: str = "JWT Bearer",
        scheme: str = "",
        secret_key: Secret | None = None,
    ):
        """
        Creates a new instance of JWTBearerAuthentication, which tries to
        obtains the identity of the user from the "Authorization" request header,
        handling JWT Bearer tokens. Only standard authorization headers starting
        with the `Bearer ` string are handled.

        Parameters
        ----------
        valid_audiences : Sequence[str]
            Sequence of acceptable audiences (aud).
        valid_issuers : Sequence[str | None]
            Sequence of acceptable issuers (iss). Required if `authority` is not
            provided. If authority is specified and issuers are not, then valid
            issuers are set as [authority].
        authority : str | None, optional
            If provided, keys are obtained from a standard well-known endpoint.
            This parameter is ignored if `keys_provider` is given.
        algorithms : Sequence[str], optional
            Sequence of acceptable algorithms. Defaults to ["RS256"] for asymmetric
            validation or ["HS256"] for symmetric validation.
        require_kid : bool, optional
            According to the specification, a key id is optional in JWK. However,
            this parameter lets control whether access tokens missing `kid` in their
            headers should be handled or rejected. By default True, thus only JWTs
            having `kid` header are accepted. Only applies to asymmetric validation.
        keys_provider : KeysProvider | None, optional
            If provided, the exact `KeysProvider` to be used when fetching JWKS for
            validation. Only applies to asymmetric validation. By default None.
        keys_url : str | None, optional
            If provided, keys are obtained from the given URL through HTTP GET.
            This parameter is ignored if `keys_provider` is given or if `secret_key`
            is provided. Only applies to asymmetric validation.
        cache_time : float, optional
            If >= 0, JWKS are cached in memory and stored for the given amount in
            seconds. By default 10800 (3 hours). Only applies to asymmetric validation.
        scheme: str
            Authentication scheme. When authentication succeeds, the identity is
            authenticated with this scheme.
        auth_mode : str, optional
            Deprecated parameter, use `scheme` instead. When authentication succeeds,
            the declared authentication mode. By default, "JWT Bearer".
        secret_key : Secret | None, optional
            If provided, enables symmetric JWT validation (HS256/HS384/HS512).
            Cannot be used together with asymmetric validation parameters.
        """
        if auth_mode != "JWT Bearer":
            # the user specified an auth_mode different than default.
            warnings.warn(
                "The auth_mode parameter is deprecated and will be removed in a "
                "future version. Use the scheme parameter instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            scheme = auth_mode
        elif scheme:
            # the user used the new parameter - good;
            auth_mode = scheme
        self._validator: BaseJWTValidator
        self.logger = get_logger()

        # Validate mutual exclusivity
        if secret_key and (authority or keys_provider or keys_url):
            raise TypeError(
                "Cannot specify both secret_key (symmetric) and "
                "authority/keys_provider/keys_url (asymmetric) parameters. "
                "Use separate instances for different validation methods."
            )

        # Determine validation mode and set appropriate defaults
        if secret_key:
            # Symmetric validation
            if not algorithms:
                algorithms = ["HS256"]

            if valid_issuers is None:
                raise TypeError("Specify valid issuers.")

            # Validate that only symmetric algorithms are specified
            invalid_algorithms = [
                algo for algo in algorithms if not algo.startswith("HS")
            ]
            if invalid_algorithms:
                raise TypeError(
                    f"When using secret_key, only HS* algorithms are supported. "
                    f"Invalid algorithms: {invalid_algorithms}"
                )

            self._validator = SymmetricJWTValidator(
                valid_issuers=valid_issuers,
                valid_audiences=valid_audiences,
                secret_key=secret_key.get_value(),
                algorithms=algorithms,
            )
        else:
            # Asymmetric validation
            if not algorithms:
                algorithms = ["RS256"]

            if authority and not valid_issuers:
                valid_issuers = [authority]

            if not authority and not valid_issuers:
                raise TypeError("Specify either an authority or valid issuers.")

            assert valid_issuers is not None

            # Validate that only asymmetric algorithms are specified
            invalid_algorithms = [
                algo
                for algo in algorithms
                if not (algo.startswith("RS") or algo.startswith("ES"))
            ]
            if invalid_algorithms:
                raise TypeError(
                    f"When using asymmetric validation, only RS*/ES* algorithms are "
                    f"supported. Invalid algorithms: {invalid_algorithms}."
                )

            self._validator = AsymmetricJWTValidator(
                authority=authority,
                algorithms=algorithms,
                require_kid=require_kid,
                keys_provider=keys_provider,
                keys_url=keys_url,
                valid_issuers=valid_issuers,
                valid_audiences=valid_audiences,
                cache_time=cache_time,
            )

        self.auth_mode = auth_mode
        self._scheme = scheme or auth_mode
        self._validator.logger = self.logger

    async def authenticate(self, context: Request) -> Identity | None:
        authorization_value = context.get_first_header(b"Authorization")

        if not authorization_value:
            return None

        if not authorization_value.startswith(b"Bearer "):
            self.logger.debug(
                "Invalid Authorization header, not starting with 'Bearer ', "
                "the header is ignored."
            )
            return None

        token = authorization_value[7:].decode()

        try:
            decoded = await self._validator.validate_jwt(token)
        except ExpiredAccessToken:
            # Common scenario
            return None
        except (InvalidAccessToken, InvalidTokenError) as exc:
            self.logger.debug(
                "JWT Bearer - access token not valid for this configuration: %s",
                str(exc),
            )
            # Raise a dedicated exception to keep track of the event
            raise InvalidCredentialsError(context.original_client_ip)
        else:
            return Identity(decoded, self.scheme)

    @property
    def scheme(self) -> str:
        return self._scheme.replace(" ", "")
